package org.wso2.carbon.application.deployer;

import org.osgi.framework.Bundle;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.wso2.carbon.utils.CarbonUtils;

import java.util.Dictionary;
import java.util.Enumeration;
import java.util.zip.ZipFile;
import java.util.zip.ZipEntry;
import java.io.*;

public class AppDeployerUtils {

    private static String APP_UNZIP_DIR;

    static {
        String javaTempDir = System.getProperty("java.io.tmpdir");
        APP_UNZIP_DIR = javaTempDir.endsWith(File.separator) ? javaTempDir + "carbon-apps" :
                javaTempDir + File.separator + "carbon-apps";
        createDir(APP_UNZIP_DIR);
    }

    private static final Log log = LogFactory.getLog(AppDeployerUtils.class);

    /**
     * Check whether the given bundle is a project artifact. If yes, return the app name. If no,
     * return null
     *
     * @param bundle - bundle to check
     * @return - app name
     */
    public static String getProjectArtifactName(Bundle bundle) {
        Dictionary dictionary = bundle.getHeaders();
        if (dictionary != null) {

            // Iterate through the headers and find the WSO2-Project-Artifact custom header
            for (Enumeration e = dictionary.keys(); e.hasMoreElements();) {

                String headerKey = (String) e.nextElement();
                if (AppDeployerConstants.WSO2_APP_NAME_HEADER.equals(headerKey)) {

                    // retireve the header value
                    String headerValue = (String) dictionary.get(headerKey);
                    if (headerValue != null) {
                        return headerValue;
                    }
                }
            }
        }
        return null;
    }

    /**
     * Checking whether a bundle contains the WSO2-Application-Deployer header
     *
     * @param bundle - input bundle
     * @return - if found header - true, else - false
     */
    public static boolean isAppDeployer(Bundle bundle) {
        Dictionary dictionary = bundle.getHeaders();
        if (dictionary != null) {

            // Iterate through the headers and find the WSO2-Project-Artifact custom header
            for (Enumeration e = dictionary.keys(); e.hasMoreElements();) {

                String headerKey = (String) e.nextElement();
                if (AppDeployerConstants.WSO2_APP_DEPLOYER_HEADER.equals(headerKey)) {
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Finds the carbon repository location
     *
     * @return - repo path
     */
    public static String getAxis2Repo() {
        String carbonRepo = System.getProperty("axis2.repo");
        if (carbonRepo == null) {
            carbonRepo = CarbonUtils.getCarbonRepositoryDir() +
                    File.separator + "deployment" + File.separator + "server";
        }
        return carbonRepo;
    }

    /**
     * Copy the artifact file at the fromPath to toPath
     *
     * @param fromPath - path at which the read the file to copy
     * @param toPath   - path to which the file should be copied
     */
    public static void copyFile(String fromPath, String toPath) {
        File in = new File(fromPath);
        if (!in.exists()) {
            log.error("Artifact file not found at : " + fromPath);
            return;
        }

        File out = new File(toPath);
        FileInputStream fis = null;
        FileOutputStream fos = null;
        try {
            fis = new FileInputStream(in);
            fos = new FileOutputStream(out);
            byte[] buf = new byte[10240];
            int i;
            while ((i = fis.read(buf)) != -1) {
                fos.write(buf, 0, i);
            }
        } catch (Exception e) {
            log.error("Error occured while copying artifact : " + fromPath, e);
        } finally {
            try {
                if (fis != null) fis.close();
                if (fos != null) fos.close();
            } catch (IOException e) {
                log.error("Error occured while closing the streams", e);
            }
        }
    }

    /**
     * Computes the application artifact file path when the bundle is given
     *
     * @param b - App artifact as an OSGi bundle
     * @return - App file path
     */
    public static String getArchivePathFromBundle(Bundle b) {
        //compute app file path
        String[] temp = b.getLocation().split("/");
        return AppDeployerUtils.getCarbonPluginLocation() + File.separator + temp[temp.length - 1];
    }

    /**
     * Finds repo/components path
     *
     * @return - path
     */
    public static String getCarbonComponentRepo() {
        return CarbonUtils.getCarbonOSGIRepoDir();
    }

    /**
     * Finds repo/synapse-config path
     *
     * @return - path
     */
    public static String getSynapseRepo() {
        return CarbonUtils.getCarbonConfigDirPath() + File.separator + "synapse-config";
    }

    /**
     * Finds repo/carbon-apps path
     *
     * @return - path
     */
    public static String getApplicationLocation() {
        return getAxis2Repo() + File.separator + "carbon-apps";
    }

    /**
     * Finds dropins location
     *
     * @return - path
     */
    public static String getCarbonPluginLocation() {
        return getCarbonComponentRepo() + File.separator + "dropins";
    }

    /**
     * Finds teh extension of a given file
     *
     * @param fileName - name of the file
     * @return - extension
     */
    public static String getFileExtension(String fileName) {
        int index = fileName.lastIndexOf('.');
        return fileName.substring(index + 1);
    }

    /**
     * Extract the Carbon application at the provided path to the java temp dir. Return the
     * extracted location
     *
     * @param appJarPath - Absolute path of the Carbon application jar file
     * @return - extracted location
     */
    public static String extractCarbonApp(String appJarPath) {
        Enumeration entries;
        ZipFile zipFile;

        String fileName = appJarPath.substring(appJarPath.lastIndexOf('/') + 1);
        String dest = APP_UNZIP_DIR + File.separator + System.currentTimeMillis() +
                fileName + File.separator;
        createDir(dest);

        try {
            zipFile = new ZipFile(appJarPath);
            entries = zipFile.entries();

            while (entries.hasMoreElements()) {
                ZipEntry entry = (ZipEntry) entries.nextElement();

                // we don't need to copy the META-INF dir
                if (entry.getName().startsWith("META-INF/")) {
                    continue;
                }

                // if the entry is a directory, create a new dir
                if (entry.isDirectory()) {
                    (new File(dest + entry.getName())).mkdir();
                    continue;
                }
                // if the entry is a file, write the file
                copyInputStream(zipFile.getInputStream(entry),
                        new BufferedOutputStream(new FileOutputStream(dest + entry.getName())));
            }
            zipFile.close();
        } catch (IOException ioe) {
            log.error("Error while extracting Carbon Application : " + fileName);
        }
        return dest;
    }

    public static void createDir(String path) {
        File temp = new File(path);
        if (!temp.exists()) {
            temp.mkdir();
        }
    }

    private static void copyInputStream(InputStream in, OutputStream out)
            throws IOException {
        byte[] buffer = new byte[40960];
        int len;

        while ((len = in.read(buffer)) >= 0)
            out.write(buffer, 0, len);

        in.close();
        out.close();
    }
}
