package org.wso2.carbon.application.deployer;

import org.osgi.framework.BundleContext;
import org.osgi.framework.Bundle;
import org.wso2.carbon.application.deployer.persistance.CarbonAppPersistanceManager;
import org.wso2.carbon.application.deployer.config.ApplicationConfiguration;
import org.wso2.carbon.application.deployer.schedule.ApplicationListener;
import org.wso2.carbon.application.deployer.schedule.Scheduler;
import org.wso2.carbon.application.deployer.schedule.SchedulerTask;
import org.wso2.carbon.application.deployer.schedule.DeploymentIterator;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.util.ArrayList;
import java.io.File;

/**
 * Manages all app deployment aspects.
 */
public class ApplicationManager {

    private static final Log log = LogFactory.getLog(ApplicationManager.class);

    private BundleContext bundleContext;

    private AppDeployerBundleListener appDeployerBundleListener = null;

    private ArrayList<AppDeploymentHandler> appDeploymentHandlers;

    private ArrayList<AppUndeploymentHandler> appUndeploymentHandlers;

    private ArrayList<CarbonApplication> carbonApps;

    //default handler is always there
    private static int initialHandlers = 1;
    private int handlerCount = 0;


    /**
     * Constructor initializes public instances
     * @param bundleContext - BundleContext of this bundle
     * @throws Exception - error on init
     */
    public ApplicationManager(BundleContext bundleContext) throws Exception {
        this.bundleContext = bundleContext;
        this.carbonApps = new ArrayList<CarbonApplication>();
        appDeploymentHandlers = new ArrayList<AppDeploymentHandler>();
        appUndeploymentHandlers = new ArrayList<AppUndeploymentHandler>();

        //load the existing Carbon apps from registry
        CarbonAppPersistanceManager.loadApps(this);

        initialHandlers += findInitialHandlers();
        registerDeploymentHandler(new DefaultAppDeployer());
    }

    /**
     * All app deployers register their deployers throgh this method
     * @param handler - app deployer which implements the AppDeploymentHandler interface
     */
    public synchronized void registerDeploymentHandler(AppDeploymentHandler handler) {
        appDeploymentHandlers.add(handler);
        handlerCount++;

        //check whether we have already initialized..
        if (initialHandlers == -1) {
            return;
        }

        if (handlerCount == initialHandlers) {
            //this is the initial server startup time, deploy apps which comes from OSGi bundles
            deployInitialCarbonApps();

            //deploy all initial apps which are not as OSGi bundles. these are from files
            ApplicationListener appListener = new ApplicationListener(this);
            appListener.findApplications(true);

            //enable hot deployment for apps
            Scheduler appScheduler = new Scheduler();
            appScheduler.schedule(new SchedulerTask(appListener), new DeploymentIterator());

            //initialHandlers count is no longer useful
            initialHandlers = -1;
        }
    }

    /**
     * All app undeployers register their instances throgh this method
     * @param handler - app undeployer which implements the AppUndeploymentHandler interface
     */
    public synchronized void registerUndeploymentHandler(AppUndeploymentHandler handler) {
        appUndeploymentHandlers.add(handler);
    }

    /**
     * Unregister the specified handler if it is already regitered
     * @param handler - input deployer handler
     */
    public synchronized void unregisterDeploymentHandler(AppDeploymentHandler handler) {
        if (appDeploymentHandlers.contains(handler)) {
            appDeploymentHandlers.remove(handler);
        }
    }

    /**
     * Unregister the specified handler if it is already regitered
     * @param handler - input undeployer handler
     */
    public synchronized void unregisterUndeploymentHandler(AppUndeploymentHandler handler) {
        if (appUndeploymentHandlers.contains(handler)) {
            appUndeploymentHandlers.remove(handler);
        }
    }

    /**
     * On server start up, deploy all existing apps
     */
    public void deployInitialCarbonApps() {
        try {
            String appName;
            for (Bundle b : bundleContext.getBundles()) {
                appName = AppDeployerUtils.getProjectArtifactName(b);
                if (appName != null) {
                    deployCarbonApp(AppDeployerUtils.getArchivePathFromBundle(b), true);
                }
            }
            //register AppDeployerBundleListener
            appDeployerBundleListener = AppDeployerBundleListener.getInstance(this);
            bundleContext.addBundleListener(appDeployerBundleListener);

        } catch (Exception e) {
            log.error("Failed to activate Carbon Application Deployer", e);
        }

    }

    /**
     * Deploy a sigle app by iterating all handlers and sending apps throgh the handler
     * chain..
     * @param archPath - carbon app archive path
     * @param isStartup - indicates whether this is server startup or not
     * @throws Exception - error on registry actions
     */
    public void deployCarbonApp(String archPath, boolean isStartup) throws Exception {
        String fileName = archPath.substring(archPath.lastIndexOf('/') + 1);
        //check whether this app already exists..
        CarbonApplication existingApp = null;
        for (CarbonApplication carbonApp : this.carbonApps) {
            if (archPath.equals(carbonApp.getAppFilePath())) {
                existingApp = carbonApp;
                break;
            }
        }

        //If the app already exists and this is the server startup, check the last updated time
        // and redeploy if needed. Return if not updated..
        if (existingApp != null) {
            /**
             * We are not supporting hot update for Carbon applications. Therefore if this is not
             * the startup, don't try to redeploy an existing application.
             */
            if (!isStartup) {
                return;
            }
            File file = new File(archPath);
            if (file.exists()) {
                Long lastUpdated = file.lastModified();
                String lastUpdatedFromRegistry =
                        CarbonAppPersistanceManager.getLastUpdatedTime(existingApp.getAppName());
                if (lastUpdatedFromRegistry != null &&
                        new Long(lastUpdatedFromRegistry).equals(lastUpdated)) {
                    return;
                } else {
                    // we are going to do a hot update for the application
                    log.warn("Carbon Application : " + fileName + " has been updated. Removing" +
                            " the existing application and redeploying...");
                    // Call the undeployer handler chain
                    for (AppUndeploymentHandler handler : appUndeploymentHandlers) {
                        handler.undeployArtifacts(existingApp);
                    }
                    this.carbonApps.remove(existingApp);
                    CarbonAppPersistanceManager.deleteApplication(existingApp.getAppName());
                }
            }
        }

        log.info("Deploying Carbon Application : " + fileName);

        CarbonApplication currentApp = new CarbonApplication();
        currentApp.setAppFilePath(archPath);
        String extractedPath = AppDeployerUtils.extractCarbonApp(archPath);

        ApplicationConfiguration appConfig = new ApplicationConfiguration(extractedPath +
                ApplicationConfiguration.APPLICATION_XML);
        currentApp.setExtractedPath(extractedPath);
        currentApp.setAppConfig(appConfig);

        String appName = appConfig.getAppId();
        if (appExists(appName)) {
            String msg = "Carbon Application : " + appName + " already exists. Two applications " +
                    "can't have the same Id. Deployment aborted.";
            log.error(msg);
            throw new Exception(msg);
        }

        //write all artifact resources to the registry
        CarbonAppPersistanceManager.writeAppResources(currentApp);

        //call the handler chain
        for (AppDeploymentHandler handler : appDeploymentHandlers) {
            handler.deployArtifacts(currentApp);
        }

        //If the app is not unusable, persist and add the app to app service
        if (!currentApp.isUnusable()) {
            currentApp.setAppName(appName);
            CarbonAppPersistanceManager.persistNewApp(currentApp);
            this.addCarbonApp(currentApp);
        } else {
            log.warn("Didn't find any deployable artifacts in Carbon App : " + fileName);
        }
    }

    /**
     * Undeploy the provided carbon App by sending it through the registered undeployment handler
     * chain..
     * @param carbonApp - CarbonApplication instance
     */
    public void undeployCarbonApp(CarbonApplication carbonApp) {
        log.info("Undeploying Carbon Application : " + carbonApp.getAppName());
        // Call the undeployer handler chain
        for (AppUndeploymentHandler handler : appUndeploymentHandlers) {
            handler.undeployArtifacts(carbonApp);
        }
        // Remove the app from application service objcet
        this.carbonApps.remove(carbonApp);

        // Remove the app from registry
        try {
            CarbonAppPersistanceManager.deleteApplication(carbonApp.getAppName());
        } catch (Exception e) {
            log.error("Can't delete the Application : " + carbonApp.getAppName());
        }
        // Remove the app artifact file from repository
        String appFilePath = carbonApp.getAppFilePath();
        File file = new File(appFilePath);
        if (file.exists() && !file.delete()) {
            log.error("Artifact file couldn't be deleted for Application : "
                    + carbonApp.getAppName());
        }
    }

    public AppDeployerBundleListener getBundleListener() {
        return appDeployerBundleListener;
    }

    public ArrayList<CarbonApplication> getCarbonApps() {
        return carbonApps;
    }

    public void addCarbonApp(CarbonApplication carbonApp) {
        this.carbonApps.add(carbonApp);
    }

    /**
     * Finds the number of app deployers at the carbon startup
     * @return - number of app deployers
     */
    private int findInitialHandlers() {
        int handlers = 0;
        for (Bundle b : bundleContext.getBundles()) {
            if (AppDeployerUtils.isAppDeployer(b)) {
                handlers++;
            }
        }
        return handlers;
    }

    /**
     * Check whether there is an already existing Carbon application with the given name
     *
     * @param newAppName - name of the new app
     * @return - true if exits
     */
    private boolean appExists(String newAppName) {
        for (CarbonApplication carbonApp : this.carbonApps) {
            if (newAppName.equals(carbonApp.getAppName())) {
                return true;
            }
        }
        return false;
    }
}
