package org.wso2.carbon.application.deployer.config;

import org.wso2.carbon.CarbonException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.OMAttribute;
import org.apache.axiom.om.impl.builder.StAXOMBuilder;

import javax.xml.stream.XMLStreamException;
import javax.xml.namespace.QName;
import java.util.List;
import java.util.Iterator;
import java.util.ArrayList;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.File;

/**
 * This is the runtime representation of the application.xml file.
 */
public class ApplicationConfiguration {

    private static final Log log = LogFactory.getLog(ApplicationConfiguration.class);

    public static final String APPLICATION_XML = "application.xml";
    public static final String ID = "Id";
    public static final String TITLE = "Title";
    public static final String VERSION = "Version";
    public static final String COMPONENTS = "Components";
    public static final String APPLICATION_NS = "http://products.wso2.org/carbon";

    private OMElement documentElement;

    private String appId;
    private String appTitle;
    private String appVersion;
    private List<Component> components;

    public ApplicationConfiguration(String appXmlPath) throws CarbonException {
        components = new ArrayList<Component>();

        File f = new File(appXmlPath);
        if (!f.exists()) {
            log.error("application.xml file not found at : " + appXmlPath);
        }
        InputStream xmlInputStream;
        try {
            xmlInputStream = new FileInputStream(f);
            documentElement = new StAXOMBuilder(xmlInputStream).getDocumentElement();
            readElements();
        } catch (FileNotFoundException e) {
            handleException("application.xml File cannot be loaded from " + appXmlPath, e);
        } catch (XMLStreamException e) {
            handleException("Problem in parsing the application.xml file ", e);
        }
    }

    public List<Component> getComponents() {
        return components;
    }

    public String getAppId() {
        return appId;
    }

    public String getAppTitle() {
        return appTitle;
    }

    public String getAppVersion() {
        return appVersion;
    }

    private void readElements() throws CarbonException {
        String temp;
        // read application attributes
        appId = readAttribute(documentElement, ID);
        appTitle = readAttribute(documentElement, TITLE);
        appVersion = readAttribute(documentElement, VERSION);

        OMElement componentsElement = documentElement
                .getFirstChildWithName(new QName(APPLICATION_NS, COMPONENTS));
        if (componentsElement == null) {
            handleException("application.xml is invalid. No Components element found");
            return;
        }

        Iterator itr = componentsElement.getChildrenWithLocalName(Component.COMPONENT);
        while (itr.hasNext()) {
            OMElement compElement = (OMElement) itr.next();
            Component comp = new Component();

            // read Id
            temp = readChildText(compElement, Component.ID);
            if (temp == null || temp.length() == 0) {
                log.error("Invalid Component in application.xml, Id not found");
            }
            comp.setId(temp);

            // read type
            temp = readChildText(compElement, Component.TYPE);
            if (temp == null || temp.length() == 0) {
                log.error("Invalid Component in application.xml, Type not found");
            }
            comp.setType(temp);

            // read other properties which are not compulsary
            comp.setArtifactName(readChildText(compElement, Component.DEPLOYABLE_ARTIFACT));
            comp.setVersion(readChildText(compElement, Component.VERSION));

            // read Resources element
            OMElement resElement = compElement
                    .getFirstChildWithName(new QName(APPLICATION_NS, Component.RESOURCES));

            if (resElement != null) {
                // read Item elements under Resources
                Iterator itemItr = resElement.getChildrenWithLocalName(Component.ITEM);
                while (itemItr.hasNext()) {
                    OMElement itemElement = (OMElement) itemItr.next();
                    comp.addResource(readChildText(itemElement, Component.PATH),
                            readChildText(itemElement, Component.CONTENT));
                }

                // read Collection elements under Resources
                Iterator collectionItr = resElement.getChildrenWithLocalName(Component.COLLECTION);
                while (collectionItr.hasNext()) {
                    OMElement collElement = (OMElement) collectionItr.next();
                    comp.addCollection(readChildText(collElement, Component.PATH),
                            readChildText(collElement, Component.DIRECTORY));
                }

                // read Association elements under Resources
                Iterator associationItr = resElement
                        .getChildrenWithLocalName(Component.ASSOCIATION);
                while (associationItr.hasNext()) {
                    OMElement assoElement = (OMElement) associationItr.next();
                    Association association = new Association();
                    association.setSourcePath(readChildText(assoElement, Component.SOURCE_PATH));
                    association.setTargetPath(readChildText(assoElement, Component.TARGET_PATH));
                    association.setType(readChildText(assoElement, Component.TYPE));
                    comp.addAssociation(association);
                }
            }

            components.add(comp);
        }
    }

    private String readChildText(OMElement element, String property) {
        if (element == null) {
            return null;
        }
        OMElement temp = element.getFirstChildWithName(new QName(APPLICATION_NS, property));
        if (temp != null) {
            return temp.getText();                
        }
        return null;
    }

    private String readAttribute(OMElement element, String attName) {
        if (element == null) {
            return null;
        }
        OMAttribute temp = element.getAttribute(new QName(attName));
        if (temp != null) {
            return temp.getAttributeValue();                
        }
        return null;
    }

    private void handleException(String msg) throws CarbonException {
        log.error(msg);
        throw new CarbonException(msg);
    }

    private void handleException(String msg, Exception e) throws CarbonException {
        log.error(msg, e);
        throw new CarbonException(msg, e);
    }

}
