/*
 * Copyright (c) 2007, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.carbon.registry.core;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.dataobjects.ResourceDO;
import org.wso2.carbon.registry.core.jdbc.utils.Transaction;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.*;

/**
 * The default registry implementation of the Collection interface.
 */
public class CollectionImpl extends ResourceImpl implements Collection {

    private static Log log = LogFactory.getLog(CollectionImpl.class);

    /**
     * The number of children in this collection.
     */
    protected int childCount;

    /**
     * The default constructor of the CollectionImpl, Create an empty collection with no children.
     */
    public CollectionImpl() {
        childCount = -1;
    }

    /**
     * Construct a collection with the provided children paths.
     *
     * @param paths the children paths.
     */
    public CollectionImpl(String[] paths) {
        String[] temp = fixPaths(paths);
        content = temp;
        childCount = temp.length;
    }

    /**
     * Construct a collection with the provided path and the resource data object.
     *
     * @param path       the path of the collection.
     * @param resourceDO the resource data object.
     */
    public CollectionImpl(String path, ResourceDO resourceDO) {
        super(path, resourceDO);
        childCount = -1;
    }

    /**
     * Implementation for the setContent. Here the content should always be a array of strings which
     * corresponding to the children paths.
     *
     * @param content array of strings which corresponding to the children paths.
     *
     * @throws RegistryException throws if the operation fails.
     */
    public void setContent(Object content) throws RegistryException {
        if (content == null) {
            return;
        }
        // note that string contents are allowed in collection to support custom generated UIs.
        if (content instanceof String[] ||
                content instanceof Resource[] ||
                content instanceof String) {
            super.setContent(content);
            return;
        }
        throw new IllegalArgumentException("Invalid content for collection. " +
                "Content of type " + content.getClass().toString() +
                " is not allowed for collections.");
    }

    /**
     * Set the resource content without marking the collection as updated.Here the content should
     * always be a array of strings which corresponding to the children paths.
     *
     * @param content array of strings which corresponding to the children paths.
     *
     * @throws RegistryException throws if the operation fails.
     */
    public void setContentWithNoUpdate(Object content) throws RegistryException {
        if (content == null) {
            return;
        }
        // note that string contents are allowed in collection to support custom generated UIs.
        if (content instanceof String[] ||
                content instanceof Resource[] ||
                content instanceof String) {
            super.setContentWithNoUpdate(content);
            return;
        }
        throw new IllegalArgumentException("Invalid content for collection. " +
                "Content of type " + content.getClass().toString() +
                " is not allowed for collections.");
    }

    /**
     * Method to return the children.
     *
     * @return an array of children paths.
     * @throws RegistryException throws if the operation fail.
     */
    public String[] getChildren() throws RegistryException {
        if (getContent() instanceof String[]) {
            return fixPaths((String[])getContent());
        } else {
            return new String[0];
        }
    }

    /**
     * Method to return the paths of the selected range of children.
     *
     * @param start   the starting number of children.
     * @param pageLen the number of entries to retrieve.
     *
     * @return an array of paths of the selected range of children.
     * @throws RegistryException throws if the operation fail.
     */
    public String[] getChildren(int start, int pageLen) throws RegistryException {
        setSessionInformation();
        try {
            if (content == null) {

                String[] childPaths = new String[0];

                if (Transaction.isStarted()) {

                    childPaths = resourceDAO.
                            getChildren(this, start, pageLen, Transaction.getConnection());

                } else {

                    Connection conn = null;
                    boolean transactionSucceeded = false;
                    try {
                        conn = dataSource.getConnection();

                        // If a managed connection already exists, use that instead of a new connection.
                        Connection temp = Transaction.getManagedRegistryConnection(conn);
                        if (temp != null) {
                            conn.close();
                            conn = temp;
                        }
                        if (conn.getTransactionIsolation() != Connection.TRANSACTION_READ_COMMITTED) {
                            conn.setTransactionIsolation(Connection.TRANSACTION_READ_COMMITTED);
                        }
                        conn.setAutoCommit(false);

                        childPaths = resourceDAO.getChildren(this, start, pageLen, conn);
                        transactionSucceeded = true;
                    } catch (SQLException e) {

                        String msg =
                                "Failed to get the child paths " + pageLen + " child paths from " +
                                        start + " of resource " + getPath() + ". " + e.getMessage();
                        log.error(msg, e);
                        throw new RegistryException(msg, e);

                    } finally {
                        if (transactionSucceeded) {
                            try {
                                conn.commit();
                            } catch (SQLException e) {
                                log.error("Failed to commit the database connection used in " +
                                        "getting child paths of the collection " + getPath());
                            }
                        } else {
                            try {
                                conn.rollback();
                            } catch (SQLException e) {
                                log.error("Failed to rollback the database connection used in " +
                                        "getting child paths of the collection " + getPath());
                            }
                        }
                        if (conn != null) {
                            try {
                                conn.close();
                            } catch (SQLException e) {
                                log.error("Failed to close the database connection used in " +
                                        "getting child paths of the collection " + getPath());
                            }
                        }
                    }
                }

                return fixPaths(childPaths);

            } else {

                if (content instanceof String[]) {

                    List<String> childPathList = new ArrayList<String>();
                    String childNodes[] = (String[]) content;
                    int limit = start + pageLen;
                    if (start > childNodes.length) {
                        return new String[0];
                    }
                    if (limit > childNodes.length) {
                        limit = childNodes.length;
                    }
                    for (int i = start; i < limit; i++) {
                        String childNode = childNodes[i];
                        childPathList.add(childNode);
                    }
                    return fixPaths(childPathList.toArray(new String[childPathList.size()]));
                }
            }
            return new String[0];
        } finally {
            clearSessionInformation();
        }
    }

    /**
     * Method to return the the number of children.
     *
     * @return the number of children.
     * @throws RegistryException throws if the operation fail.
     */
    public int getChildCount() throws RegistryException {

        try {
            setSessionInformation();

            if (childCount != -1) {
                return childCount;

            } else if (content != null && content instanceof String[]) {

                String[] childPaths = (String[]) content;
                return fixPaths(childPaths).length;

            } else {

                if (Transaction.isStarted()) {

                    childCount = resourceDAO.getChildCount(this, Transaction.getConnection());

                } else {

                    Connection conn = null;
                    boolean transactionSucceeded = false;
                    try {
                        conn = dataSource.getConnection();

                        // If a managed connection already exists, use that instead of a new connection.
                        Connection temp = Transaction.getManagedRegistryConnection(conn);
                        if (temp != null) {
                            conn.close();
                            conn = temp;
                        }
                        if (conn.getTransactionIsolation() != Connection.TRANSACTION_READ_COMMITTED) {
                            conn.setTransactionIsolation(Connection.TRANSACTION_READ_COMMITTED);
                        }
                        conn.setAutoCommit(false);

                        childCount = resourceDAO.getChildCount(this, conn);
                        transactionSucceeded = true;

                    } catch (SQLException e) {

                        String msg = "Failed to get the child count of resource " +
                                getPath() + ". " + e.getMessage();
                        log.error(msg, e);
                        throw new RegistryException(msg, e);

                    } finally {
                        if (transactionSucceeded) {
                            try {
                                conn.commit();
                            } catch (SQLException e) {
                                log.error("Failed to commit the database connection used in " +
                                        "getting child count of the collection " + getPath());
                            }
                        } else {
                            try {
                                conn.rollback();
                            } catch (SQLException e) {
                                log.error("Failed to rollback the database connection used in " +
                                        "getting child count of the collection " + getPath());
                            }
                        }
                        if (conn != null) {
                            try {
                                conn.close();
                            } catch (SQLException e) {
                                log.error("Failed to close the database connection used in " +
                                        "getting child count of collection " + getPath());
                            }
                        }
                    }
                }
            }

            return childCount;

        } finally {
            clearSessionInformation();
        }
    }

    /**
     * Method to set the child count.
     *
     * @param count the child count.
     */
    public void setChildCount(int count) {
        childCount = count;
    }


    /**
     * Collection's content is a string array, which contains paths of its children. These paths are
     * loaded on demand to increase performance. It is recommended to use {@link #getChildren()}
     * method to get child paths of a collection, which provides pagination. Calling this method
     * will load all child paths.
     *
     * @return String array of child paths.
     * @throws RegistryException On any error.
     */
    public Object getContent() throws RegistryException {
        setSessionInformation();
        try {

            if (content == null) {
                if (Transaction.isStarted()) {
                    resourceDAO.fillChildren(this, 0, -1, Transaction.getConnection());
                } else {

                    Connection conn = null;
                    boolean transactionSucceeded = false;
                    try {
                        conn = dataSource.getConnection();

                        // If a managed connection already exists, use that instead of a new connection.
                        Connection temp = Transaction.getManagedRegistryConnection(conn);
                        if (temp != null) {
                            conn.close();
                            conn = temp;
                        }
                        if (conn.getTransactionIsolation() != Connection.TRANSACTION_READ_COMMITTED) {
                            conn.setTransactionIsolation(Connection.TRANSACTION_READ_COMMITTED);
                        }
                        conn.setAutoCommit(false);

                        resourceDAO.fillChildren(this, 0, -1, conn);
                        transactionSucceeded = true;
                    } catch (SQLException e) {

                        String msg = "Failed to get child paths of " +
                                getPath() + ". " + e.getMessage();
                        log.error(msg, e);
                        throw new RegistryException(msg, e);

                    } finally {
                        if (transactionSucceeded) {
                            try {
                                conn.commit();
                            } catch (SQLException e) {
                                log.error("Failed to commit the database connection used in " +
                                        "getting child paths of the collection " + getPath());
                            }
                        } else {
                            try {
                                conn.rollback();
                            } catch (SQLException e) {
                                log.error("Failed to rollback the database connection used in " +
                                        "getting child paths of the collection " + getPath());
                            }
                        }
                        if (conn != null) {
                            try {
                                conn.close();
                            } catch (SQLException e) {
                                log.error("Failed to close the database connection opened in " +
                                        "getting the child paths of " + getPath(), e);
                            }
                        }
                    }
                }
            }
            return content;

        } finally {
            clearSessionInformation();
        }
    }

    /**
     * Method to return a shallow copy of a collection.
     *
     * @return the shallow copy of the collection.
     * @throws RegistryException throws if the operation fail.
     */
    public ResourceImpl getShallowCopy() throws RegistryException {
        CollectionImpl newCollection = new CollectionImpl();
        fillCollectionCopy(newCollection);
        return newCollection;
    }

    /**
     * Copy all the values of the current collection attribute to the passed collection.
     *
     * @param collection the collection to get all the current collection attribute copied.
     *
     * @throws RegistryException throws if the operation fail.
     */
    public void fillCollectionCopy(CollectionImpl collection) throws RegistryException {
        super.fillResourceCopy(collection);
        collection.setChildCount(this.childCount);
    }

    /**
     * Method to fix duplicated entries in a collection's child paths.
     * @param paths the collection's child paths.
     * @return the distinct set of children.
     */
    @SuppressWarnings("ManualArrayToCollectionCopy")
    protected String[] fixPaths(String[] paths) {
        Set<String> temp = new LinkedHashSet<String>();
        // We want to make sure that each element is added one after the other in the exact order
        // that they were passed in.
        for (String path : paths) {
            temp.add(path);
        }
        return temp.toArray(new String[temp.size()]);
    }
}
