/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.dao;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.Association;
import org.wso2.carbon.registry.core.RegistryConstants;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.DatabaseConstants;
import org.wso2.carbon.registry.core.jdbc.utils.Transaction;
import org.wso2.carbon.registry.core.session.CurrentSession;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;

/**
 * Data Access Object for Associations
 */
public class AssociationDAO {

    private static final Log log = LogFactory.getLog(AssociationDAO.class);

    /**
     * Method to persist an association.
     *
     * @param sourcePath      the source path
     * @param targetPath      the target path
     * @param associationType the type of the association
     *
     * @throws RegistryException if some error occurs while adding associations
     */
    public void addAssociation(String sourcePath,
                               String targetPath,
                               String associationType) throws RegistryException {
        if (isAssociationExisting(sourcePath, targetPath, associationType)) {
            // not to duplicate the associations.
            return;
        }
        Connection conn = Transaction.getConnection();

        PreparedStatement ps = null;
        try {
            String propSQL =
                    "INSERT INTO REG_ASSOCIATION (REG_SOURCEPATH, REG_TARGETPATH, " +
                            "REG_ASSOCIATION_TYPE, REG_TENANT_ID) VALUES (?,?,?,?)";
            ps = conn.prepareStatement(propSQL);
            ps.setString(1, sourcePath);
            ps.setString(2, targetPath);
            ps.setString(3, associationType);
            ps.setInt(4, CurrentSession.getTenantId());
            ps.executeUpdate();

        } catch (SQLException e) {

            String msg = "Failed to add association between resources " +
                    sourcePath + " and " + targetPath + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) {
                    ps.close();
                }
            } catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }

    // Determines whether an association exists or not.
    private boolean isAssociationExisting(String sourcePath,
                                          String targetPath,
                                          String associationType) throws RegistryException {
        boolean isExisting = false;
        Connection conn = Transaction.getConnection();
        ResultSet result = null;
        PreparedStatement ps = null;
        try {
            String propSQL =
                    "SELECT REG_SOURCEPATH FROM REG_ASSOCIATION WHERE REG_SOURCEPATH=? AND " +
                            "REG_TARGETPATH=? AND REG_ASSOCIATION_TYPE=? AND REG_TENANT_ID=?";
            ps = conn.prepareStatement(propSQL);
            ps.setString(1, sourcePath);
            ps.setString(2, targetPath);
            ps.setString(3, associationType);
            ps.setInt(4, CurrentSession.getTenantId());
            result = ps.executeQuery();

            if (result.next()) {
                isExisting = true;
            }

        } catch (SQLException e) {
            String msg = "Failed to check the existence of the association between resources " +
                    sourcePath + " and " + targetPath + " for association type " +
                    associationType + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) {
                    ps.close();
                }
                if (result != null) {
                    result.close();
                }
            } catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
        return isExisting;
    }

    /**
     * Method to remove an association.
     *
     * @param sourcePath      the source path
     * @param targetPath      the target path
     * @param associationType the type of the association
     *
     * @throws RegistryException if some error occurs while removing associations
     */
    public void removeAssociation(String sourcePath,
                                  String targetPath,
                                  String associationType) throws RegistryException {

        Connection conn = Transaction.getConnection();

        PreparedStatement ps = null;
        try {
            String propSQL =
                    "DELETE FROM REG_ASSOCIATION WHERE REG_SOURCEPATH=? AND REG_TARGETPATH=? " +
                            "AND REG_ASSOCIATION_TYPE=? AND REG_TENANT_ID=?";
            ps = conn.prepareStatement(propSQL);
            ps.setString(1, sourcePath);
            ps.setString(2, targetPath);
            ps.setString(3, associationType);
            ps.setInt(4, CurrentSession.getTenantId());
            ps.executeUpdate();

        } catch (SQLException e) {
            String msg = "Failed to remove association between resources " +
                    sourcePath + " and " + targetPath + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) {
                    ps.close();
                }
            } catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }

    /**
     * Method to get all association.
     *
     * @param resourcePath the source path
     *
     * @return the array of all associations
     * @throws RegistryException if some error occurs while getting all associations
     */
    public Association[] getAllAssociations(String resourcePath) throws RegistryException {

        Connection conn = Transaction.getConnection();
        ArrayList<Association> allAssociations;
        PreparedStatement ps = null;
        ResultSet result = null;
        try {
            String propSQL = "SELECT REG_SOURCEPATH,REG_TARGETPATH,REG_ASSOCIATION_TYPE FROM " +
                    "REG_ASSOCIATION WHERE (REG_SOURCEPATH=? OR REG_TARGETPATH=?)" +
                    " AND REG_TENANT_ID=?";
            ps = conn.prepareStatement(propSQL);
            ps.setString(1, resourcePath);
            ps.setString(2, resourcePath);
            ps.setInt(3, CurrentSession.getTenantId());
            result = ps.executeQuery();
            allAssociations = new ArrayList<Association>();

            while (result.next()) {
                Association association = new Association();
                association.setSourcePath(result.getString(DatabaseConstants.SOURCEPATH_FIELD));
                association
                        .setDestinationPath(result.getString(DatabaseConstants.TARGETPATH_FIELD));
                association.setAssociationType(
                        result.getString(DatabaseConstants.ASSOCIATION_TYPE_FIELD));
                allAssociations.add(association);
            }
        } catch (SQLException e) {
            String msg = "Failed to get all associations of resource " +
                    resourcePath + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            } catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }

        return allAssociations.toArray(new Association[allAssociations.size()]);
    }

    /**
     * Method to get all association of a given type.
     *
     * @param resourcePath    the source path
     * @param associationType the type of the association
     *
     * @return the array of associations of a given type
     * @throws RegistryException if some error occurs while getting associations of a given type
     */
    public Association[] getAllAssociationsForType(String resourcePath, String associationType)
            throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;
        ArrayList<Association> associations;
        try {
            String propSQL = "SELECT REG_SOURCEPATH, REG_TARGETPATH FROM REG_ASSOCIATION " +
                    "WHERE (REG_SOURCEPATH=? OR REG_TARGETPATH=?) " +
                    "AND REG_ASSOCIATION_TYPE=? AND REG_TENANT_ID=?";
            ps = conn.prepareStatement(propSQL);
            ps.setString(1, resourcePath);
            ps.setString(2, resourcePath);
            ps.setString(3, associationType);
            ps.setInt(4, CurrentSession.getTenantId());

            result = ps.executeQuery();
            associations = new ArrayList<Association>();

            while (result.next()) {
                Association association = new Association();
                association.setSourcePath(result.getString(DatabaseConstants.SOURCEPATH_FIELD));
                association
                        .setDestinationPath(result.getString(DatabaseConstants.TARGETPATH_FIELD));
                association.setAssociationType(associationType);
                associations.add(association);
            }
        } catch (SQLException e) {
            String msg = "Failed to get associations of type " +
                    associationType + " for resource " + resourcePath + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            } catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }

        return associations.toArray(new Association[associations.size()]);
    }

    /**
     * Method to replace all associations, when moving or renaming a resource. All the associations
     * towards a given path can be replaced by the new path by calling this method.
     *
     * @param oldPath the old resource path
     * @param newPath the new resource path
     *
     * @throws RegistryException if an error occurs while
     */
    public void replaceAssociations(String oldPath, String newPath) throws RegistryException {

        Connection conn = Transaction.getConnection();

        PreparedStatement ps2 = null;
        try {
            String sql2 = "UPDATE REG_ASSOCIATION SET REG_TARGETPATH=? WHERE " +
                    "REG_TARGETPATH=? AND REG_TENANT_ID=?";

            ps2 = conn.prepareStatement(sql2);
            ps2.setString(1, newPath);
            ps2.setString(2, oldPath);
            ps2.setInt(3, CurrentSession.getTenantId());
            ps2.executeUpdate();

        } catch (SQLException e) {

            String msg = "Failed to replace the associations of " + oldPath +
                    " by re-associating them to " + newPath + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps2 != null) {
                    ps2.close();
                }
            } catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }

    /**
     * Method to remove all associations for a given path.
     *
     * @param resourcePath the source path
     *
     * @throws RegistryException if an error occurs while
     */
    public void removeAllAssociations(String resourcePath) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "DELETE FROM REG_ASSOCIATION WHERE (REG_SOURCEPATH=? " +
                "OR REG_TARGETPATH=?) AND REG_TENANT_ID=?";

        PreparedStatement ps = null;
        try {
            ps = conn.prepareStatement(sql);
            ps.setString(1, resourcePath);
            ps.setString(2, resourcePath);
            ps.setInt(3, CurrentSession.getTenantId());
            ps.executeUpdate();

        } catch (SQLException e) {

            String msg = "Failed to remove associations of resource " +
                    resourcePath + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR +
                        ex.getMessage();
                log.error(msg, ex);
            }
        }
    }

    /**
     * Method to copy associations for a given path to a new path.
     *
     * @param fromPath the source path
     * @param toPath   the destination path
     *
     * @throws RegistryException if an error occurs while
     */
    public void copyAssociations(String fromPath, String toPath) throws RegistryException {
        PreparedStatement ps = null;
        ResultSet result = null;
        try {
            Connection conn = Transaction.getConnection();
            String propSQL =
                    "SELECT REG_TARGETPATH, REG_ASSOCIATION_TYPE FROM REG_ASSOCIATION WHERE " +
                            "REG_SOURCEPATH=? AND REG_TENANT_ID=?";
            ps = conn.prepareStatement(propSQL);
            ps.setString(1, fromPath);
            ps.setInt(2, CurrentSession.getTenantId());
            result = ps.executeQuery();
            ArrayList<Association> arrayList = new ArrayList<Association>();

            // Read all results first
            while (result.next()) {
                Association association = new Association();
                association.setSourcePath(toPath);
                association
                        .setDestinationPath(result.getString(DatabaseConstants.TARGETPATH_FIELD));
                association.setAssociationType(
                        result.getString(DatabaseConstants.ASSOCIATION_TYPE_FIELD));
                arrayList.add(association);
            }

            // Then add associations
            for (Association association : arrayList) {
                addAssociation(toPath, association.getDestinationPath(),
                        association.getAssociationType());
            }
        } catch (SQLException e) {
            String msg = "SQLException occurred during copyAssociations";
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            } catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }
}
