/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.carbon.registry.core.jdbc.queries;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.exceptions.RegistryException;

import java.util.StringTokenizer;

/**
 * class to transform the sql to be tenant-aware. Once the query is passed to the constructor this
 * will be keeping the transformed query.
 */
public class TenantAwareSQLTransformer {

    private static final Log log = LogFactory.getLog(TenantAwareSQLTransformer.class);
    private String transformedQuery = null;
    int parameterCount = 0;
    int trailingParameterCount = 0;

    /**
     * Construct the TenantAwareSQLTransformer for a given query
     *
     * @param sqlQuery the query to transform to tenant aware sql
     *
     * @throws RegistryException throws if the transformation failed.
     */
    TenantAwareSQLTransformer(String sqlQuery) throws RegistryException {
        String sqlQueryUCase = sqlQuery.toUpperCase();
        parameterCount = 0;
        int endOfFromIndex = sqlQueryUCase.indexOf("FROM");
        if (endOfFromIndex == -1) {
            String msg = "Error in parsing the query. You should have a 'FROM' token in your " +
                    "custom query";
            log.error(msg);
            throw new RegistryException(msg);
        }
        endOfFromIndex += 4;
        int startOfWhereIndex = sqlQueryUCase.indexOf("WHERE");
        int startOfOrderByIndex = sqlQueryUCase.indexOf("ORDER BY");
        boolean whereNotFound = false;
        if (startOfWhereIndex == -1) {
            // no 'WHERE'
            whereNotFound = true;
            startOfWhereIndex = sqlQueryUCase.length();
            if (startOfOrderByIndex != -1) {
                startOfWhereIndex = startOfOrderByIndex;
            }
        }
        String fromPart = sqlQuery.substring(endOfFromIndex + 1, startOfWhereIndex);
        StringTokenizer tokenizer = new StringTokenizer(fromPart, ",");

        String additionalWherePart = null;

        while (tokenizer.hasMoreElements()) {
            String token = tokenizer.nextToken();
            token = token.trim();
            token = token.replaceAll("[\t\r\n]+", " ");
            int separator = token.indexOf(' ');
            String firstPart;
            String secondPart;
            if (separator == -1) {
                firstPart = token;
                secondPart = null;
            } else {
                firstPart = token.substring(0, separator);
                secondPart = token.substring(separator + 1);
                firstPart = firstPart.trim();
                secondPart = secondPart.trim();
            }
            // now the first part contains the table name
            if (secondPart == null) {
                if (additionalWherePart == null) {
                    additionalWherePart = firstPart + ".REG_TENANT_ID=?";
                } else {
                    additionalWherePart =
                            additionalWherePart + " AND " + firstPart + ".REG_TENANT_ID=?";
                }
            } else {
                if (additionalWherePart == null) {
                    additionalWherePart = secondPart + ".REG_TENANT_ID=?";
                } else {
                    additionalWherePart =
                            additionalWherePart + " AND " + secondPart + ".REG_TENANT_ID=?";
                }
            }
            parameterCount++;
        }
        if (whereNotFound) {
            if (startOfOrderByIndex == -1) {
                transformedQuery = sqlQuery + " WHERE " + additionalWherePart;
            } else {
                String[] parts = sqlQuery.substring(startOfOrderByIndex).split("[?]");
                if (parts != null && parts.length > 1) {
                    trailingParameterCount += parts.length - 1;
                    if (sqlQuery.substring(startOfOrderByIndex).endsWith("?")) {
                        trailingParameterCount += 1;
                    }
                }
                transformedQuery = sqlQuery.substring(0, startOfOrderByIndex) + " WHERE "
                        + additionalWherePart + " " + sqlQuery.substring(startOfOrderByIndex);
            }
        } else {
            int endOfWhereIndex = startOfWhereIndex + 5;
            if (startOfOrderByIndex == -1) {
                transformedQuery = sqlQuery.substring(0, endOfWhereIndex) + " (" +
                        sqlQuery.substring(endOfWhereIndex) + ") " + " AND "
                        + additionalWherePart;
            } else {
                String[] parts = sqlQuery.substring(startOfOrderByIndex).split("[?]");
                if (parts != null && parts.length > 1) {
                    trailingParameterCount += parts.length - 1;
                    if (sqlQuery.substring(startOfOrderByIndex).endsWith("?")) {
                        trailingParameterCount += 1;
                    }
                }
                transformedQuery = sqlQuery.substring(0, endOfWhereIndex) + " (" +
                        sqlQuery.substring(endOfWhereIndex, startOfOrderByIndex) + ") " + " AND "
                        + additionalWherePart + " " + sqlQuery.substring(startOfOrderByIndex);
            }
        }
    }

    /**
     * Get the transformed query.
     *
     * @return the transformed query.
     */
    String getTransformedQuery() {
        return transformedQuery;
    }

    /**
     * Get additional parameter count. The additional number of tenant id to be set.
     *
     * @return the additional parameter count.
     */
    int getAdditionalParameterCount() {
        return parameterCount;
    }

    /**
     * Get trailing parameter count.
     *
     * @return the trailing parameter count
     */
    int getTrailingParameterCount() {
        return trailingParameterCount;
    }
}