/*
 * Copyright 2004,2005 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.carbon.server.transports.http;

import org.apache.catalina.LifecycleException;
import org.apache.catalina.connector.Connector;
import org.apache.catalina.startup.Embedded;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.server.TomcatServerCache;
import org.wso2.carbon.server.transports.Transport;
import org.wso2.carbon.server.transports.TransportParameter;
import org.wso2.carbon.utils.ServerConstants;

import java.net.InetAddress;
import java.util.List;

/**
 * Corresponds to the HTTP transport when Carbon is running standlone. Responsible for managing
 * the corresponding Tomcat connector.
 */
public class HttpTransport implements Transport {
    private static final Log log = LogFactory.getLog(HttpTransport.class);
    private boolean isStarted;
    private Connector connector;
    private int proxyPort = -1;

    public void init(List<TransportParameter> parameters) throws Exception {
        Embedded embedded = TomcatServerCache.getEmbedded();
        try {
            for (TransportParameter parameter : parameters) {
                if (parameter.getName().equals("port")) {
                    int port;
                    String portSysProp = System.getProperty("httpPort");
                    if (portSysProp != null){
                        try {
                            port = Integer.parseInt(portSysProp);
                        } catch (NumberFormatException ignored) {
                            port = Integer.parseInt(parameter.getValue());
                        }
                    } else {
                        port = Integer.parseInt(parameter.getValue());
                    }
                    connector = embedded.createConnector((InetAddress) null, port, false);
                } else if (parameter.getName().equals("proxyPort")) {
                    proxyPort = Integer.parseInt(parameter.getValue());
                } else {
                    connector.setAttribute(parameter.getName(), parameter.getValue());
                }
            }
            connector.setEnableLookups(true);
        } catch (Exception
                e) {
            embedded.removeConnector(connector);
            String msg = "Exception occurred while starting the Tomcat HTTP connector";
            log.error(msg, e);
            throw new Exception(msg, e);
        }
    }

    public void start() throws Exception {
        if (isStarted) {
            return;
        }
        if (log.isDebugEnabled()) {
            log.debug("Starting HTTP Transport...");
        }
        if (connector != null) {
            TomcatServerCache.addConnector(connector);
            connector.start();
        } else {
            throw new RuntimeException("Cannot start transport before initializing it. " +
                                       "Please  call #init() first.");
        }
        isStarted = true;
    }

    public void stop() throws Exception {
        if (!isStarted) {
            return;
        }
        if (log.isDebugEnabled()) {
            log.debug("Stopping HTTP transport...");
        }
        try {
            if (connector.isAvailable()) {
                connector.stop();
                TomcatServerCache.removeConnector(connector);
            }
            isStarted = false;
        } catch (LifecycleException e) {
            String msg = "Cannot stop Tomcat HTTP connector";
            log.error(msg, e);
            throw new Exception(msg, e);
        }
    }

    public int getPort() {
        return connector.getPort();
    }

    public int getProxyPort() {
        return proxyPort;
    }
}
