/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.user.core.authorization;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

import javax.sql.DataSource;

import net.sf.ehcache.Cache;
import net.sf.ehcache.CacheManager;
import net.sf.ehcache.Element;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.CarbonConstants;
import org.wso2.carbon.user.core.AuthorizationManager;
import org.wso2.carbon.user.core.UserCoreConstants;
import org.wso2.carbon.user.core.UserRealm;
import org.wso2.carbon.user.core.UserStoreException;
import org.wso2.carbon.user.core.claim.ClaimManager;
import org.wso2.carbon.user.core.config.RealmConfiguration;
import org.wso2.carbon.user.core.internal.UMListenerServiceComponent;
import org.wso2.carbon.user.core.listener.AuthorizationManagerListener;
import org.wso2.carbon.user.core.profile.ProfileConfigurationManager;
import org.wso2.carbon.user.core.util.DatabaseUtil;
import org.wso2.carbon.user.core.util.UserCoreUtil;

public class JDBCAuthorizationManager implements AuthorizationManager {

    private DataSource dataSource = null;
    private PermissionTree permissionTree = null;
    private CacheManager cacheManager = null;
    private UserRealm userRealm = null;
    private int tenantId;
    /**
     * The root node of the tree
     */
    private static Log log = LogFactory.getLog(JDBCAuthorizationManager.class);

    public JDBCAuthorizationManager(RealmConfiguration realmConfig, Map<String, Object> properties,
            ClaimManager claimManager, ProfileConfigurationManager profileManager, UserRealm realm,
            Integer tenantId) throws UserStoreException {
        cacheManager = new CacheManager();
        Cache cache1 = new Cache(UserCoreConstants.AUTHZ_CACHE,
                UserCoreConstants.MAX_OBJECTS_IN_CACHE, false, false, UserCoreConstants.TTL_CACHE,
                UserCoreConstants.IDLE_TIME_IN_CACHE);
        cacheManager.addCache(cache1);
        dataSource = (DataSource) properties.get(UserCoreConstants.DATA_SOURCE);
        if (dataSource == null) {
            dataSource = DatabaseUtil.createRealmDataSource(realmConfig);
            properties.put(UserCoreConstants.DATA_SOURCE, dataSource);
        }
        this.permissionTree = new PermissionTree();
        // this.realmConfig = realmConfig;
        this.userRealm = realm;
        this.tenantId = tenantId;
        if (log.isDebugEnabled()) {
            log.debug("The dataSource being used by JDBCAuthorizationManager :: "
                    + dataSource.hashCode());
        }
        this.populatePermissionTreeFromDB();
    }

    public boolean isRoleAuthorized(String roleName, String resourceId, String action)
            throws UserStoreException {

        for (AuthorizationManagerListener listener : UMListenerServiceComponent
                .getAuthorizationManagerListeners()) {
            if (!listener.isRoleAuthorized(roleName, resourceId, action, this)) {
                return false;
            }
        }

        SearchResult sr = permissionTree.getRolePermission(roleName, PermissionTreeUtil
                .actionToPermission(action), null, null, PermissionTreeUtil
                .toComponenets(resourceId));
        if (sr.getLastNodeAllowedAccess() == Boolean.TRUE) {
            return true;
        }
        return false;
    }

    public boolean isUserAuthorized(String userName, String resourceId, String action)
            throws UserStoreException {

        if (CarbonConstants.REGISTRY_SYSTEM_USERNAME.equals(userName)) {
            return true;
        }

        for (AuthorizationManagerListener listener : UMListenerServiceComponent
                .getAuthorizationManagerListeners()) {
            if (!listener.isUserAuthorized(userName, resourceId, action, this)) {
                return false;
            }
        }

        Cache cache = cacheManager.getCache(UserCoreConstants.AUTHZ_CACHE);
        Element element = cache.get(userName + resourceId + action);

        if (element != null) {
            Boolean bool = (Boolean) element.getObjectValue();
            return bool.booleanValue();
        }

        SearchResult sr = permissionTree.getUserPermission(userName, PermissionTreeUtil
                .actionToPermission(action), null, null, PermissionTreeUtil
                .toComponenets(resourceId));
        if (sr.getLastNodeAllowedAccess() == Boolean.TRUE) {
            return true;
        }

        String[] userRoles = userRealm.getUserStoreManager().getRoleListOfUser(userName);
        boolean userAllowed = false;
        List<String> allowedRoles = Arrays.asList(getAllowedRolesForResource(resourceId, action));

        for (String role : userRoles) {
            if (allowedRoles.contains(role)) {
                userAllowed = true;
                break;
            }
        }
        return userAllowed;
    }

    public String[] getAllowedRolesForResource(String resourceId, String action)
            throws UserStoreException {
        TreeNode.Permission p = PermissionTreeUtil.actionToPermission(action);
        SearchResult sr = permissionTree.getAllowedRolesForResource(null, null, p,
                PermissionTreeUtil.toComponenets(resourceId));

        return sr.getAllowedEntities().toArray(new String[sr.getAllowedEntities().size()]);
    }

    public String[] getExplicitlyAllowedUsersForResource(String resourceId, String action)
            throws UserStoreException {
        TreeNode.Permission p = PermissionTreeUtil.actionToPermission(action);
        SearchResult sr = permissionTree.getAllowedUsersForResource(null, null, p,
                PermissionTreeUtil.toComponenets(resourceId));

        return sr.getAllowedEntities().toArray(new String[sr.getAllowedEntities().size()]);
    }

    public String[] getDeniedRolesForResource(String resourceId, String action)
            throws UserStoreException {
        TreeNode.Permission p = PermissionTreeUtil.actionToPermission(action);
        SearchResult sr = permissionTree.getDeniedRolesForResource(null, null, p,
                PermissionTreeUtil.toComponenets(resourceId));
        return sr.getDeniedEntities().toArray(new String[sr.getAllowedEntities().size()]);
    }

    public String[] getExplicitlyDeniedUsersForResource(String resourceId, String action)
            throws UserStoreException {
        TreeNode.Permission p = PermissionTreeUtil.actionToPermission(action);
        SearchResult sr = permissionTree.getDeniedUsersForResource(null, null, p,
                PermissionTreeUtil.toComponenets(resourceId));
        return sr.getDeniedEntities().toArray(new String[sr.getAllowedEntities().size()]);
    }

    public String[] getAllowedUIResourcesForUser(String userName, String permissionRootPath)
            throws UserStoreException {
        List<String> lstPermissions = new ArrayList<String>();
        String[] roles = this.userRealm.getUserStoreManager().getRoleListOfUser(userName);
        TreeNode node = permissionTree.getNodeForPath(permissionRootPath);
        if (node == null) {
            throw new UserStoreException("Invalid Permission root path provided");
        }

        TreeNode permissionNode = node.getChild(CarbonConstants.UI_PERMISSION_NAME);
        if (permissionNode == null) {
            throw new UserStoreException("Invalid Permission root path provided");
        }
               
        permissionTree.getUIResourcesForRoles(roles, lstPermissions, "", null, permissionNode);
        String[] permissions = lstPermissions.toArray(new String[lstPermissions.size()]);
        String[] optimizedList = UserCoreUtil.optimizePermissions(permissions);
        return optimizedList;
    }

    public void authorizeRole(String roleName, String resourceId, String action)
            throws UserStoreException {

        for (AuthorizationManagerListener listener : UMListenerServiceComponent
                .getAuthorizationManagerListeners()) {
            if (!listener.authorizeRole(roleName, resourceId, action, this)) {
                return;
            }
        }

        if (resourceId == null || action == null) {
            log.error("Invalid data provided at authorization code");
            throw new UserStoreException("Invalid data provided");
        }
        addAuthorizationForRole(roleName, resourceId, action, UserCoreConstants.ALLOW);
    }

    public void denyRole(String roleName, String resourceId, String action)
            throws UserStoreException {

        for (AuthorizationManagerListener listener : UMListenerServiceComponent
                .getAuthorizationManagerListeners()) {
            if (!listener.denyRole(roleName, resourceId, action, this)) {
                return;
            }
        }

        if (resourceId == null || action == null) {
            log.error("Invalid data provided at authorization code");
            throw new UserStoreException("Invalid data provided");
        }
        addAuthorizationForRole(roleName, resourceId, action, UserCoreConstants.DENY);
    }

    public void authorizeUser(String userName, String resourceId, String action)
            throws UserStoreException {

        for (AuthorizationManagerListener listener : UMListenerServiceComponent
                .getAuthorizationManagerListeners()) {
            if (!listener.authorizeUser(userName, resourceId, action, this)) {
                return;
            }
        }

        if (resourceId == null || action == null) {
            log.error("Invalid data provided at authorization code");
            throw new UserStoreException("Invalid data provided");
        }
        addAuthorizationForUser(userName, resourceId, action, UserCoreConstants.ALLOW);
    }

    public void denyUser(String userName, String resourceId, String action)
            throws UserStoreException {

        for (AuthorizationManagerListener listener : UMListenerServiceComponent
                .getAuthorizationManagerListeners()) {
            if (!listener.denyUser(userName, resourceId, action, this)) {
                return;
            }
        }

        if (resourceId == null || action == null) {
            log.error("Invalid data provided at authorization code");
            throw new UserStoreException("Invalid data provided");
        }
        addAuthorizationForUser(userName, resourceId, action, UserCoreConstants.DENY);
    }

    public void clearResourceAuthorizations(String resourceId) throws UserStoreException {

        for (AuthorizationManagerListener listener : UMListenerServiceComponent
                .getAuthorizationManagerListeners()) {
            if (!listener.clearResourceAuthorizations(resourceId, this)) {
                return;
            }
        }

        Connection dbConnection = null;
        try {
            dbConnection = getDBConnection();
            DatabaseUtil.updateDatabase(dbConnection,
                    DBConstants.ON_DELETE_PERMISSION_UM_ROLE_PERMISSIONS_SQL, resourceId, tenantId);
            DatabaseUtil.updateDatabase(dbConnection,
                    DBConstants.ON_DELETE_PERMISSION_UM_USER_PERMISSIONS_SQL, resourceId, tenantId);
            DatabaseUtil.updateDatabase(dbConnection, DBConstants.DELETE_PERMISSION_SQL,
                    resourceId, tenantId);
            permissionTree.clearResourceAuthorizations(resourceId);
            dbConnection.commit();
        } catch (SQLException e) {
            log.error("Error! " + e.getMessage(), e);
            throw new UserStoreException("Error! " + e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection);
        }
    }

    public void clearRoleAuthorization(String roleName, String resourceId, String action)
            throws UserStoreException {

        for (AuthorizationManagerListener listener : UMListenerServiceComponent
                .getAuthorizationManagerListeners()) {
            if (!listener.clearRoleAuthorization(roleName, resourceId, action, this)) {
                return;
            }
        }

        Connection dbConnection = null;
        try {
            dbConnection = getDBConnection();
            DatabaseUtil.updateDatabase(dbConnection, DBConstants.DELETE_ROLE_PERMISSION_SQL,
                    roleName, resourceId, action, tenantId, tenantId);
            permissionTree.clearRoleAuthorization(roleName, resourceId, action);
            dbConnection.commit();
        } catch (SQLException e) {
            log.error("Error! " + e.getMessage(), e);
            throw new UserStoreException("Error! " + e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection);
        }
    }

    public void clearUserAuthorization(String userName, String resourceId, String action)
            throws UserStoreException {

        for (AuthorizationManagerListener listener : UMListenerServiceComponent
                .getAuthorizationManagerListeners()) {
            if (!listener.clearUserAuthorization(userName, resourceId, action, this)) {
                return;
            }
        }

        Connection dbConnection = null;
        PreparedStatement prepStmt = null;
        try {
            dbConnection = getDBConnection();
            int permissionId = this.getPermissionId(dbConnection, resourceId, action);
            if (permissionId == -1) {
                this.addPermissionId(dbConnection, resourceId, action);
                permissionId = this.getPermissionId(dbConnection, resourceId, action);
            }
            DatabaseUtil.updateDatabase(dbConnection, DBConstants.DELETE_USER_PERMISSION_SQL,
                    userName, resourceId, action, tenantId, tenantId);
            permissionTree.clearUserAuthorization(userName, resourceId, action);
            dbConnection.commit();
        } catch (SQLException e) {
            log.error("Error! " + e.getMessage(), e);
            throw new UserStoreException("Error! " + e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, prepStmt);
        }
    }

    public void clearRoleActionOnAllResources(String roleName, String action)
            throws UserStoreException {

        for (AuthorizationManagerListener listener : UMListenerServiceComponent
                .getAuthorizationManagerListeners()) {
            if (!listener.clearRoleActionOnAllResources(roleName, action, this)) {
                return;
            }
        }

        Connection dbConnection = null;
        PreparedStatement prepStmt = null;
        try {
            dbConnection = getDBConnection();
            permissionTree.clearRoleAuthorization(roleName, action);
            DatabaseUtil.updateDatabase(dbConnection,
                    DBConstants.DELETE_ROLE_PERMISSIONS_BASED_ON_ACTION, roleName, action,
                    tenantId, tenantId);
            dbConnection.commit();
        } catch (SQLException e) {
            log.error("Error! " + e.getMessage(), e);
            throw new UserStoreException("Error! " + e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, prepStmt);
        }
    }

    public void clearRoleAuthorization(String roleName) throws UserStoreException {

        for (AuthorizationManagerListener listener : UMListenerServiceComponent
                .getAuthorizationManagerListeners()) {
            if (!listener.clearRoleAuthorization(roleName, this)) {
                return;
            }
        }

        Connection dbConnection = null;
        PreparedStatement prepStmt = null;
        try {
            dbConnection = getDBConnection();
            permissionTree.clearRoleAuthorization(roleName);
            DatabaseUtil.updateDatabase(dbConnection,
                    DBConstants.ON_DELETE_ROLE_DELETE_PERMISSION_SQL, roleName, tenantId);
            dbConnection.commit();
        } catch (SQLException e) {
            log.error("Error! " + e.getMessage(), e);
            throw new UserStoreException("Error! " + e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, prepStmt);
        }
    }

    public void clearUserAuthorization(String userName) throws UserStoreException {

        for (AuthorizationManagerListener listener : UMListenerServiceComponent
                .getAuthorizationManagerListeners()) {
            if (!listener.clearUserAuthorization(userName, this)) {
                return;
            }
        }

        Connection dbConnection = null;
        PreparedStatement prepStmt = null;
        try {
            dbConnection = getDBConnection();
            permissionTree.clearUserAuthorization(userName);
            DatabaseUtil.updateDatabase(dbConnection,
                    DBConstants.ON_DELETE_USER_DELETE_PERMISSION_SQL, userName, tenantId);
            dbConnection.commit();
        } catch (SQLException e) {
            log.error("Error! " + e.getMessage(), e);
            throw new UserStoreException("Error! " + e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, prepStmt);
        }

    }

    private void addAuthorizationForRole(String roleName, String resourceId, String action,
            short allow) throws UserStoreException {
        Connection dbConnection = null;
        PreparedStatement prepStmt = null;
        try {
            dbConnection = getDBConnection();
            int permissionId = this.getPermissionId(dbConnection, resourceId, action);
            if (permissionId == -1) {
                this.addPermissionId(dbConnection, resourceId, action);
                permissionId = this.getPermissionId(dbConnection, resourceId, action);
            }
            DatabaseUtil.updateDatabase(dbConnection, DBConstants.DELETE_ROLE_PERMISSION_SQL,
                    roleName, resourceId, action, tenantId, tenantId);
            DatabaseUtil.updateDatabase(dbConnection, DBConstants.ADD_ROLE_PERMISSION_SQL,
                    permissionId, roleName, allow, tenantId);

            if (allow == UserCoreConstants.ALLOW) {
                permissionTree.authorizeRoleInTree(roleName, resourceId, action);
            } else {
                permissionTree.denyRoleInTree(roleName, resourceId, action);
            }
            dbConnection.commit();
        } catch (SQLException e) {
            log.error("Error! " + e.getMessage(), e);
            throw new UserStoreException("Error! " + e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, prepStmt);
        }
    }

    private void addAuthorizationForUser(String userName, String resourceId, String action,
            short allow) throws UserStoreException {
        Connection dbConnection = null;
        PreparedStatement prepStmt = null;
        try {
            dbConnection = getDBConnection();
            int permissionId = this.getPermissionId(dbConnection, resourceId, action);
            if (permissionId == -1) {
                this.addPermissionId(dbConnection, resourceId, action);
                permissionId = this.getPermissionId(dbConnection, resourceId, action);
            }
            DatabaseUtil.updateDatabase(dbConnection, DBConstants.DELETE_USER_PERMISSION_SQL,
                    userName, resourceId, action, tenantId, tenantId);
            DatabaseUtil.updateDatabase(dbConnection, DBConstants.ADD_USER_PERMISSION_SQL,
                    permissionId, userName, allow, tenantId);
            if (allow == UserCoreConstants.ALLOW) {
                permissionTree.authorizeUserInTree(userName, resourceId, action);
            } else {
                permissionTree.denyUserInTree(userName, resourceId, action);
            }
            dbConnection.commit();
        } catch (SQLException e) {
            log.error("Error! " + e.getMessage(), e);
            throw new UserStoreException("Error! " + e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, prepStmt);
        }
    }

    private int getPermissionId(Connection dbConnection, String resourceId, String action)
            throws UserStoreException {
        PreparedStatement prepStmt = null;
        ResultSet rs = null;
        int value = -1;
        try {
            prepStmt = dbConnection.prepareStatement(DBConstants.GET_PERMISSION_ID_SQL);
            prepStmt.setString(1, resourceId);
            prepStmt.setString(2, action);
            prepStmt.setInt(3, tenantId);

            rs = prepStmt.executeQuery();
            if (rs.next()) {
                value = rs.getInt(1);
            }
            return value;
        } catch (SQLException e) {
            log.error("Error! " + e.getMessage(), e);
            throw new UserStoreException("Error! " + e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(null, rs, prepStmt);
        }
    }

    private void addPermissionId(Connection dbConnection, String resourceId, String action)
            throws UserStoreException {
        PreparedStatement prepStmt = null;
        try {
            prepStmt = dbConnection.prepareStatement(DBConstants.ADD_PERMISSION_SQL);
            prepStmt.setString(1, resourceId);
            prepStmt.setString(2, action);
            prepStmt.setInt(3, tenantId);
            int count = prepStmt.executeUpdate();
            if (log.isDebugEnabled()) {
                log.debug("Executed querry is " + DBConstants.ADD_PERMISSION_SQL
                        + " and number of updated rows :: " + count);
            }
        } catch (SQLException e) {
            log.error("Error! " + e.getMessage(), e);
            throw new UserStoreException("Error! " + e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(null, prepStmt);
        }
    }

    private Connection getDBConnection() throws SQLException {
        Connection dbConnection = dataSource.getConnection();
        dbConnection.setAutoCommit(false);
        return dbConnection;
    }

    private void populatePermissionTreeFromDB() throws UserStoreException {
        ResultSet rs = null;
        PreparedStatement prepStmt1 = null;
        PreparedStatement prepStmt2 = null;
        Connection dbConnection = null;
        try {
            dbConnection = getDBConnection();
            // Populating role permissions
            prepStmt1 = dbConnection.prepareStatement(DBConstants.GET_EXISTING_ROLE_PERMISSIONS);
            prepStmt1.setInt(1, tenantId);
            prepStmt1.setInt(2, tenantId);
            rs = prepStmt1.executeQuery();

            while (rs.next()) {
                short allow = rs.getShort(3);
                if (allow == UserCoreConstants.ALLOW) {
                    permissionTree.authorizeRoleInTree(rs.getString(1), rs.getString(2), rs
                            .getString(4));
                } else {
                    permissionTree
                            .denyRoleInTree(rs.getString(1), rs.getString(2), rs.getString(4));
                }
            }

            // Populating user permissions
            prepStmt2 = dbConnection.prepareStatement(DBConstants.GET_EXISTING_USER_PERMISSIONS);
            prepStmt2.setInt(1, tenantId);
            prepStmt2.setInt(2, tenantId);
            rs = prepStmt2.executeQuery();

            while (rs.next()) {
                short allow = rs.getShort(3);
                if (allow == UserCoreConstants.ALLOW) {
                    permissionTree.authorizeUserInTree(rs.getString(1), rs.getString(2), rs
                            .getString(4));
                } else {
                    permissionTree
                            .denyUserInTree(rs.getString(1), rs.getString(2), rs.getString(4));
                }

            }

        } catch (SQLException e) {
            e.printStackTrace();
            log.error(e.getMessage(), e);
            throw new UserStoreException(
                    "Error loading authorizations. Please check the database. Error message is "
                            + e.getMessage(), e);
        } finally {
            DatabaseUtil.closeAllConnections(dbConnection, rs, prepStmt1, prepStmt2);
        }
    }

    public int getTenantId() throws UserStoreException {
        return tenantId;
    }
}
