/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.user.core.authorization;

import java.util.BitSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.wso2.carbon.CarbonConstants;
import org.wso2.carbon.user.core.UserStoreException;

public class PermissionTree {

    protected TreeNode root;

    /**
     * On the server startup, all permissions are populated from the DB and the
     * permission tree is built in memory..
     * 
     * @throws UserStoreException
     *             - SQL exceptions
     */

    public PermissionTree() {
        root = new TreeNode("/");
    }

    void authorizeUserInTree(String userName, String resourceId, String action) {
        SearchResult sr = getNode(root, PermissionTreeUtil.toComponenets(resourceId));
        if (sr.getUnprocessedPaths() != null) {
            List<String> paths = sr.getUnprocessedPaths();
            TreeNode tn = sr.getLastNode().create(paths);
            tn.authorizeUser(userName, PermissionTreeUtil.actionToPermission(action));

        } else {
            sr.getLastNode().authorizeUser(userName, PermissionTreeUtil.actionToPermission(action));
        }
    }

    void denyUserInTree(String userName, String resourceId, String action)
            throws UserStoreException {
        SearchResult sr = getNode(root, PermissionTreeUtil.toComponenets(resourceId));
        if (sr.getUnprocessedPaths() != null) {
            List<String> paths = sr.getUnprocessedPaths();
            TreeNode tn = sr.getLastNode().create(paths);
            tn.denyUser(userName, PermissionTreeUtil.actionToPermission(action));
        } else {
            sr.getLastNode().denyUser(userName, PermissionTreeUtil.actionToPermission(action));
        }
    }

    void authorizeRoleInTree(String roleName, String resourceId, String action)
            throws UserStoreException {

        SearchResult sr = getNode(root, PermissionTreeUtil.toComponenets(resourceId));
        if (sr.getUnprocessedPaths() != null) {
            List<String> paths = sr.getUnprocessedPaths();
            TreeNode tn = sr.getLastNode().create(paths);
            tn.authorizeRole(roleName, PermissionTreeUtil.actionToPermission(action));

        } else {
            sr.getLastNode().authorizeRole(roleName, PermissionTreeUtil.actionToPermission(action));
        }
    }

    void denyRoleInTree(String roleName, String resourceId, String action)
            throws UserStoreException {

        SearchResult sr = getNode(root, PermissionTreeUtil.toComponenets(resourceId));
        if (sr.getUnprocessedPaths() != null) {
            List<String> paths = sr.getUnprocessedPaths();
            TreeNode tn = sr.getLastNode().create(paths);
            tn.denyRole(roleName, PermissionTreeUtil.actionToPermission(action));

        } else {
            sr.getLastNode().denyRole(roleName, PermissionTreeUtil.actionToPermission(action));
        }
    }

    /**
     * Find permission for given 'role' starting from the TreeNode 'node' and
     * using the path segments given in the List as pathParts
     * 
     * @param role
     *            the role for which the permissions are searched
     * @param p
     *            the permission checked
     * @param sr
     *            the SearchResult that merges the permissions as the tree is
     *            traversed
     * @param node
     *            the current node
     * @param pathParts
     *            the list of path segments to traverse
     * @return the final SearchResult that merges the permissions
     */
    SearchResult getRolePermission(String role, TreeNode.Permission p, SearchResult sr,
            TreeNode node, List<String> pathParts) {

        if (node == null) {
            node = root;
        }

        if (sr == null) {
            sr = new SearchResult();
        }

        Boolean currentNodeAllows = node.isRoleAuthorized(role, p);
        if (currentNodeAllows == Boolean.TRUE) {
            sr.setLastNodeAllowedAccess(Boolean.TRUE);
        } else if (currentNodeAllows == Boolean.FALSE) {
            sr.setLastNodeAllowedAccess(Boolean.FALSE);
        }

        if (pathParts == null || pathParts.isEmpty()) {
            sr.setLastNode(node);
            sr.setUnprocessedPaths(null);
            return sr;

        } else {

            String key = pathParts.get(0);
            if (key != null && key.length() > 0) {
                TreeNode child = node.getChild(key);
                if (child != null) {
                    pathParts.remove(0);
                    return getRolePermission(role, p, sr, child, pathParts);
                }
            }
            sr.setLastNode(node);
            return sr;
        }
    }

    /**
     * Find permission for given 'user' starting from the TreeNode 'node' and
     * using the path segments given in the List as pathParts
     * 
     * @param user
     *            the user for which the permissions are searched
     * @param p
     *            the permission checked
     * @param sr
     *            the SearchResult that merges the permissions as the tree is
     *            traversed
     * @param node
     *            the current node
     * @param pathParts
     *            the list of path segments to traverse
     * @return the final SearchResult that merges the permissions
     */
    SearchResult getUserPermission(String user, TreeNode.Permission p, SearchResult sr,
            TreeNode node, List<String> pathParts) {

        if (node == null) {
            node = root;
        }
        if (sr == null) {
            sr = new SearchResult();
        }

        Boolean currentNodeAllows = node.isUserAuthorized(user, p);
        if (currentNodeAllows == Boolean.TRUE) {
            sr.setLastNodeAllowedAccess(Boolean.TRUE);
        } else if (currentNodeAllows == Boolean.FALSE) {
            sr.setLastNodeAllowedAccess(Boolean.FALSE);
        }

        if (pathParts == null || pathParts.isEmpty()) {
            sr.setLastNode(node);
            sr.setUnprocessedPaths(null);
            return sr;

        } else {

            String key = pathParts.get(0);
            if (key != null && key.length() > 0) {
                TreeNode child = node.getChild(key);
                if (child != null) {
                    pathParts.remove(0);
                    return getUserPermission(user, p, sr, child, pathParts);
                }
            }
            sr.setLastNode(node);
            return sr;
        }
    }

    /**
     * Find the allowed Users for a given resource by traversing the whole
     * pemission tree.
     * 
     * @param sr
     *            - search result to contain allowed users
     * @param node
     *            - current node
     * @param p
     *            - permission
     * @param pathParts
     *            - list of path segments to traverse
     * @return - search result with allowed users
     */

    SearchResult getAllowedUsersForResource(SearchResult sr, TreeNode node, TreeNode.Permission p,
            List<String> pathParts) {

        if (node == null) {
            node = root;
        }
        if (sr == null) {
            sr = new SearchResult();
        }

        /**
         * Add allowed users of the current node to our list in the sr
         */
        Map<String, BitSet> allowUsers = node.getUserAllowPermissions();
        for (String user : allowUsers.keySet()) {
            BitSet bs = allowUsers.get(user);
            if (bs.get(p.ordinal())) {
                if (!sr.getAllowedEntities().contains(user)) {
                    sr.getAllowedEntities().add(user);
                }
            }
        }

        /**
         * Remove denied users of the current node from our list in the sr
         */
        Map<String, BitSet> denyUsers = node.getUserDenyPermissions();
        for (String user : denyUsers.keySet()) {
            BitSet bs = denyUsers.get(user);
            if (bs.get(p.ordinal()) && sr.getAllowedEntities().contains(user)) {
                sr.getAllowedEntities().remove(user);
            }
        }

        if (pathParts == null || pathParts.isEmpty()) {
            sr.setLastNode(node);
            sr.setUnprocessedPaths(null);
            return sr;
        } else {
            String key = pathParts.get(0);
            if (key != null && key.length() > 0) {
                TreeNode child = node.getChild(key);
                if (child != null) {
                    pathParts.remove(0);
                    return getAllowedUsersForResource(sr, child, p, pathParts);
                }
            }
            sr.setLastNode(node);
            return sr;
        }
    }

    /**
     * Find the allowed Roles for a given resource by traversing the whole
     * pemission tree.
     * 
     * @param sr
     *            - search result to contain allowed roles
     * @param node
     *            - current node
     * @param p
     *            - permission
     * @param pathParts
     *            - list of path segments to traverse
     * @return - search result with allowed roles
     */

    SearchResult getAllowedRolesForResource(SearchResult sr, TreeNode node, TreeNode.Permission p,
            List<String> pathParts) {
        if (node == null) {
            node = root;
        }

        if (sr == null) {
            sr = new SearchResult();
        }

        /**
         * Add allowed roles of the current node to our list in the sr
         */
        Map<String, BitSet> allowRoles = node.getRoleAllowPermissions();
        for (String role : allowRoles.keySet()) {
            BitSet bs = allowRoles.get(role);
            if (bs.get(p.ordinal())) {
                if (!sr.getAllowedEntities().contains(role)) {
                    sr.getAllowedEntities().add(role);
                }
            }
        }

        /**
         * Remove denied roles of the current node from our list in the sr
         */
        Map<String, BitSet> denyRoles = node.getRoleDenyPermissions();
        for (String role : denyRoles.keySet()) {
            BitSet bs = denyRoles.get(role);
            if (bs.get(p.ordinal()) && sr.getAllowedEntities().contains(role)) {
                sr.getAllowedEntities().remove(role);
            }
        }

        if (pathParts == null || pathParts.isEmpty()) {
            sr.setLastNode(node);
            sr.setUnprocessedPaths(null);
            return sr;
        } else {
            String key = pathParts.get(0);
            if (key != null && key.length() > 0) {
                TreeNode child = node.getChild(key);
                if (child != null) {
                    pathParts.remove(0);
                    return getAllowedRolesForResource(sr, child, p, pathParts);
                }
            }
            sr.setLastNode(node);
            return sr;
        }
    }

    /**
     * Find the denied Roles for a given resource by traversing the whole
     * pemission tree.
     * 
     * @param sr
     *            - search result to contain denied roles
     * @param node
     *            - current node
     * @param p
     *            - permission
     * @param pathParts
     *            - list of path segments to traverse
     * @return - search result with denied roles
     */
    SearchResult getDeniedRolesForResource(SearchResult sr, TreeNode node, TreeNode.Permission p,
            List<String> pathParts) {
        if (sr == null) {
            sr = new SearchResult();
        }

        if (node == null) {
            node = root;
        }

        /**
         * Add denied roles of the current node to our list in the sr
         */
        Map<String, BitSet> denyRoles = node.getRoleDenyPermissions();
        for (String role : denyRoles.keySet()) {
            BitSet bs = denyRoles.get(role);
            if (bs.get(p.ordinal())) {
                if (!sr.getDeniedEntities().contains(role)) {
                    sr.getDeniedEntities().add(role);
                }
            }
        }

        /**
         * Remove allowed roles of the current node from our list in the sr
         */
        Map<String, BitSet> allowRoles = node.getRoleAllowPermissions();
        for (String role : allowRoles.keySet()) {
            BitSet bs = allowRoles.get(role);
            if (bs.get(p.ordinal()) && sr.getDeniedEntities().contains(role)) {
                sr.getDeniedEntities().remove(role);
            }
        }

        if (pathParts == null || pathParts.isEmpty()) {
            sr.setLastNode(node);
            sr.setUnprocessedPaths(null);
            return sr;
        } else {
            String key = pathParts.get(0);
            if (key != null && key.length() > 0) {
                TreeNode child = node.getChild(key);
                if (child != null) {
                    pathParts.remove(0);
                    return getDeniedRolesForResource(sr, child, p, pathParts);
                }
            }
            sr.setLastNode(node);
            return sr;
        }
    }

    /**
     * Find the denied users for a given resource by traversing the whole
     * pemission tree.
     * 
     * @param sr
     *            - search result to contain denied users
     * @param node
     *            - current node
     * @param p
     *            - permission
     * @param pathParts
     *            - list of path segments to traverse
     * @return - search result with denied users
     */
    SearchResult getDeniedUsersForResource(SearchResult sr, TreeNode node, TreeNode.Permission p,
            List<String> pathParts) {
        if (sr == null) {
            sr = new SearchResult();
        }

        if (node == null) {
            node = root;
        }

        /**
         * Add denied users of the current node to our list in the sr
         */
        Map<String, BitSet> denyUsers = node.getUserDenyPermissions();
        for (String user : denyUsers.keySet()) {
            BitSet bs = denyUsers.get(user);
            if (bs.get(p.ordinal())) {
                if (!sr.getDeniedEntities().contains(user)) {
                    sr.getDeniedEntities().add(user);
                }
            }
        }

        /**
         * Remove allowed users of the current node from our list in the sr
         */
        Map<String, BitSet> allowUsers = node.getUserAllowPermissions();
        for (String role : allowUsers.keySet()) {
            BitSet bs = allowUsers.get(role);
            if (bs.get(p.ordinal()) && sr.getDeniedEntities().contains(role)) {
                sr.getDeniedEntities().remove(role);
            }
        }

        if (pathParts == null || pathParts.isEmpty()) {
            sr.setLastNode(node);
            sr.setUnprocessedPaths(null);
            return sr;
        } else {
            String key = pathParts.get(0);
            if (key != null && key.length() > 0) {
                TreeNode child = node.getChild(key);
                if (child != null) {
                    pathParts.remove(0);
                    return getDeniedUsersForResource(sr, child, p, pathParts);
                }
            }
            sr.setLastNode(node);
            return sr;
        }
    }

    /**
     * Find a node on the tree, starting from the given nodes, and using the
     * list of path segments
     * 
     * @param node
     *            the start node to begin the search
     * @param pathParts
     *            a List of path segments - i.e. collection/resource names
     * @return the result as a SearchResult
     * @see SearchResult
     */
    protected SearchResult getNode(TreeNode node, List<String> pathParts) {
        if (pathParts == null || pathParts.isEmpty()) {
            return new SearchResult(node, null);

        } else {
            String key = pathParts.get(0);
            if (key != null && key.length() > 0) {
                TreeNode child = node.getChild(key);
                if (child != null) {
                    pathParts.remove(0);
                    if (!pathParts.isEmpty()) {
                        return getNode(child, pathParts);
                    } else {
                        return new SearchResult(child, null);
                    }
                }
            }
            return new SearchResult(node, pathParts);
        }
    }

    /**
     * Clear all role authorization based on action starting from the given node
     * 
     * @param node
     *            the start node to begin the search
     * @param
     * @return the result as a SearchResultpathParts a List of
     * @see SearchResult
     */
    void clearRoleAuthorization(String roleName, TreeNode node, TreeNode.Permission p) {
        Map<String, BitSet> allowRoles = node.getRoleAllowPermissions();
        Map<String, BitSet> denyRoles = node.getRoleDenyPermissions();

        BitSet bs = allowRoles.get(roleName);
        if (bs != null) {
            bs.clear(p.ordinal());
        }

        bs = denyRoles.get(roleName);
        if (bs != null) {
            bs.clear(p.ordinal());
        }

        Map<String, TreeNode> childMap = node.getChildren();
        if (childMap != null && childMap.size() > 0) {
            Iterator<TreeNode> ite = childMap.values().iterator();
            while (ite.hasNext()) {
                TreeNode childNode = ite.next();
                clearRoleAuthorization(roleName, childNode, p);
            }
        }
    }

    /**
     * Clear all role authorization based on action starting from the given node
     * 
     * @param node
     *            the start node to begin the search
     * @param
     * @return the result as a SearchResultpathParts a List of
     * @see SearchResult
     */
    void clearRoleAuthorization(String roleName, TreeNode node) {
        Map<String, BitSet> allowRoles = node.getRoleAllowPermissions();
        Map<String, BitSet> denyRoles = node.getRoleDenyPermissions();

        BitSet bs = allowRoles.get(roleName);
        if (bs != null) {
            allowRoles.remove(roleName);
        }

        bs = denyRoles.get(roleName);
        if (bs != null) {
            denyRoles.remove(roleName);
        }

        Map<String, TreeNode> childMap = node.getChildren();
        if (childMap != null && childMap.size() > 0) {
            Iterator<TreeNode> ite = childMap.values().iterator();
            while (ite.hasNext()) {
                TreeNode childNode = ite.next();
                clearRoleAuthorization(roleName, childNode);
            }
        }
    }

    public void clearRoleAuthorization(String roleName, String resourceId, String action)
            throws UserStoreException {

        SearchResult sr = getNode(root, PermissionTreeUtil.toComponenets(resourceId));
        if (sr.getUnprocessedPaths() == null) {
            TreeNode.Permission p = PermissionTreeUtil.actionToPermission(action);

            Map<String, BitSet> allowRoles = sr.getLastNode().getRoleAllowPermissions();
            BitSet bs = allowRoles.get(roleName);
            if (bs != null) {
                bs.clear(p.ordinal());
            }

            Map<String, BitSet> denyRoles = sr.getLastNode().getRoleDenyPermissions();
            bs = denyRoles.get(roleName);
            if (bs != null) {
                bs.clear(p.ordinal());
            }
        }
    }

    public void clearRoleAuthorization(String roleName, String action) throws UserStoreException {
        TreeNode.Permission p = PermissionTreeUtil.actionToPermission(action);
        clearRoleAuthorization(roleName, root, p);
    }

    public void clearRoleAuthorization(String roleName) throws UserStoreException {
        clearRoleAuthorization(roleName, root);
    }

    void clearUserAuthorization(String userName, TreeNode node) {
        Map<String, BitSet> allowUsers = node.getUserAllowPermissions();
        Map<String, BitSet> denyUsers = node.getUserDenyPermissions();

        BitSet bs = allowUsers.get(userName);
        if (bs != null) {
            allowUsers.remove(userName);
        }

        bs = denyUsers.get(userName);
        if (bs != null) {
            denyUsers.remove(userName);
        }

        Map<String, TreeNode> childMap = node.getChildren();
        if (childMap != null && childMap.size() > 0) {
            Iterator<TreeNode> ite = childMap.values().iterator();
            while (ite.hasNext()) {
                TreeNode childNode = ite.next();
                clearUserAuthorization(userName, childNode);
            }
        }
    }

    public void clearUserAuthorization(String userName) throws UserStoreException {
        clearUserAuthorization(userName, root);
    }

    public void clearUserAuthorization(String userName, String resourceId, String action)
            throws UserStoreException {

        SearchResult sr = getNode(root, PermissionTreeUtil.toComponenets(resourceId));
        if (sr.getUnprocessedPaths() == null || sr.getUnprocessedPaths().isEmpty()) {
            TreeNode.Permission p = PermissionTreeUtil.actionToPermission(action);

            Map<String, BitSet> allowUsers = sr.getLastNode().getUserAllowPermissions();
            BitSet bs = allowUsers.get(userName);
            if (bs != null) {
                bs.clear(p.ordinal());
            }

            Map<String, BitSet> denyUsers = sr.getLastNode().getUserDenyPermissions();
            bs = denyUsers.get(userName);
            if (bs != null) {
                bs.clear(p.ordinal());
            }
        }
    }

    public void clearResourceAuthorizations(String resourceId) throws UserStoreException {

        SearchResult sr = getNode(root, PermissionTreeUtil.toComponenets(resourceId));
        if (sr.getUnprocessedPaths() == null) {
            sr.getLastNode().getUserAllowPermissions().clear();
            sr.getLastNode().getUserDenyPermissions().clear();
            sr.getLastNode().getRoleAllowPermissions().clear();
            sr.getLastNode().getRoleDenyPermissions().clear();
        }
    }

    /**
     * This method is only used for UI permissions
     * 
     * @param sr
     *            - search result to contain allowed users
     * @param node
     *            - current node
     * @param p
     *            - permission
     * @param pathParts
     *            - list of path segments to traverse
     * @return - search result with allowed users
     */

    void getUIResourcesForRoles(String[] roles, List<String> resources, String path,
            TreeNode.Permission p, TreeNode node) {

        String currentPath = path + "/" + node.getName();
        
        if (p == null) {
            p = PermissionTreeUtil.actionToPermission(CarbonConstants.UI_PERMISSION_ACTION);
        }

        Map<String, BitSet> bsAllowed = node.getRoleAllowPermissions();
        for (String role : roles) {
            BitSet bs = bsAllowed.get(role);
            if (bs != null && bs.get(p.ordinal())) {
                resources.add(currentPath);
                return;
            }
        }

        Map<String, TreeNode> children = node.getChildren();
        Iterator<TreeNode> nodeItr = children.values().iterator();
        while (nodeItr.hasNext()) {
            TreeNode child = nodeItr.next();
            if (child != null) {
                getUIResourcesForRoles(roles, resources, currentPath, p, child);
            }
        }
    }

    TreeNode getNodeForPath(String path) {
        List<String> paths = PermissionTreeUtil.toComponenets(path);
        TreeNode node = root;
        for (Iterator<String> ite = paths.iterator(); ite.hasNext();) {
            String name = ite.next();
            node = node.getChild(name);
            if (node == null) {
                return null;
            }
        }
        return node;
    }

}
