package org.wso2.carbon.user.core.authorization;

import java.util.BitSet;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * A node in the Tree structure used to maintain hierarchical security permissions. The growth
 * of the tree is on the order of explicit permission statements, and not on the number of
 * resources whose permissions are maintained.
 *
 */
public class TreeNode {

    public static enum Permission { GET, ADD, DELETE, EDIT, LOGIN, MAN_CONFIG, MAN_LC_CONFIG, MAN_SEC, UP_SERV,
        MAN_SERV, MAN_MEDIA, MON_SYS, DEL_ID, AUTHORIZE, INV_SER, UI_EXECUTE}

    /** The name of the node - For the Registry, this would be the name of a Collection/Rsource */
    private String name;
    /** The children of this node - maintained on a Map by the names */
    private Map<String, TreeNode> children = new HashMap<String, TreeNode>();
    /** Explicit allow permission for specific users */
    private Map<String, BitSet> userAllowPermissions = new HashMap<String, BitSet>();
    /** Explicit deny permission for specific users */
    private Map<String, BitSet> userDenyPermissions  = new HashMap<String, BitSet>();
    /** Explicit allow permission for specific roles */
    private Map<String, BitSet> roleAllowPermissions = new HashMap<String, BitSet>();
    /** Explicit deny permission for specific roles */
    private Map<String, BitSet> roleDenyPermissions  = new HashMap<String, BitSet>();

    /**
     * Constructor
     * @param name the name of the TreeNode
     */
    TreeNode(String name) {
        this.name = name;
    }

    /**
     * Get the child by the given name
     * @param name name of the child node
     * @return the child with the given name, or null
     */
    TreeNode getChild(String name) {
        return children.get(name);
    }

    /**
     * Is the 'user' authorized for the given permission p on this node?
     * @param user the name of the user
     * @param p the permission
     * @return Boolean.TRUE if authorized, Boolean.FALSE if not
     */
    public Boolean isUserAuthorized(String user, Permission p) {
        BitSet bsAlow = userAllowPermissions.get(user);
        BitSet bsDeny = userDenyPermissions.get(user);
        if (bsAlow == null && bsDeny == null) {
            return null;
        } else if (bsDeny != null && bsDeny.get(p.ordinal())) {
            return Boolean.FALSE;
        } else if (bsAlow != null && bsAlow.get(p.ordinal())) {
            return Boolean.TRUE;
        }
        return null;
    }

    /**
     * Is the 'role' authorized for the given permission p on this node?
     * @param role the name of the role
     * @param p the permission
     * @return Boolean.TRUE if authorized, Boolean.FALSE if not
     */
    public Boolean isRoleAuthorized(String role, Permission p) {
        BitSet bsAlow = roleAllowPermissions.get(role);
        BitSet bsDeny = roleDenyPermissions.get(role);
        if (bsAlow == null && bsDeny == null) {
            return null;
        } else if (bsDeny != null && bsDeny.get(p.ordinal())) {
            return Boolean.FALSE;
        } else if (bsAlow != null && bsAlow.get(p.ordinal())) {
            return Boolean.TRUE;
        }
        return null;
    }

    /**
     * Grant explicit authorization to the 'user' on this node for permission p
     * @param user the user who is granted authorization
     * @param p the permission granted
     */
    public void authorizeUser(String user, Permission p) {
        BitSet bsAllow = userAllowPermissions.get(user);
        if (bsAllow == null) {
            bsAllow = new BitSet();
            bsAllow.set(p.ordinal());
            userAllowPermissions.put(user, bsAllow);
        } else {
            bsAllow.set(p.ordinal());
        }

        BitSet bsDeny = userDenyPermissions.get(user);
        if (bsDeny != null) {
            bsDeny.clear(p.ordinal());
        }
    }

    /**
     * Grant explicit authorization to the 'role' on this node for permission p
     * @param role the role that is granted authorization
     * @param p the permission granted
     */
    public void authorizeRole(String role, Permission p) {
        BitSet bsAllow = roleAllowPermissions.get(role);
        if (bsAllow == null) {
            bsAllow = new BitSet();
            bsAllow.set(p.ordinal());
            roleAllowPermissions.put(role, bsAllow);
        } else {
            bsAllow.set(p.ordinal());
        }

        BitSet bsDeny = roleDenyPermissions.get(role);
        if (bsDeny != null) {
            bsDeny.clear(p.ordinal());
        }
    }

    /**
     * Deny explicit authorization to the 'user' on this node for permission p
     * @param user the user that is denied authorization
     * @param p the permission denied
     */
    public void denyUser(String user, Permission p) {
        BitSet bsDeny = userDenyPermissions.get(user);
        if (bsDeny == null) {
            bsDeny = new BitSet();
            bsDeny.set(p.ordinal());
            userDenyPermissions.put(user, bsDeny);
        } else {
            bsDeny.set(p.ordinal());
        }

        BitSet bsAllow = userAllowPermissions.get(user);
        if (bsAllow != null) {
            bsAllow.clear(p.ordinal());
        }
    }

    /**
     * Deny explicit authorization to the 'role' on this node for permission p
     * @param role the role that is denied authorization
     * @param p the permission denied
     */
    public void denyRole(String role, Permission p) {
        BitSet bsDeny = roleDenyPermissions.get(role);
        if (bsDeny == null) {
            bsDeny = new BitSet();
            bsDeny.set(p.ordinal());
            roleDenyPermissions.put(role, bsDeny);
        } else {
            bsDeny.set(p.ordinal());
        }

        BitSet bsAllow = roleAllowPermissions.get(role);
        if (bsAllow != null) {
            bsAllow.clear(p.ordinal());
        }
    }

    /**
     * Create the tree structure for the given paths array of nodes
     * @param paths an array of hierarchical nodes to be created, in-order
     * @return the reference to the lowest decendent created
     */
    public TreeNode create(List<String> paths) {
        if (paths != null && !paths.isEmpty()) {
            String childName = paths.get(0);
            TreeNode tn = new TreeNode(childName);
            children.put(childName, tn);
            paths.remove(0);
            if (!paths.isEmpty()) {
                return tn.create(paths);
            } else {
                return tn;
            }
        } else {
            return this;
        }
    }

    /**
     * The name of the node
     * @return node name
     */
    public String getName() {
        return name;
    }

    /**
     * The children of the node as a Map keyed by the name
     * @return the children as a Map
     */
    public Map<String, TreeNode> getChildren() {
        return children;
    }

    //-------- getters --------
    public Map<String, BitSet> getUserAllowPermissions() {
        return userAllowPermissions;
    }

    public Map<String, BitSet> getUserDenyPermissions() {
        return userDenyPermissions;
    }

    public Map<String, BitSet> getRoleAllowPermissions() {
        return roleAllowPermissions;
    }

    public Map<String, BitSet> getRoleDenyPermissions() {
        return roleDenyPermissions;
    }
}
