/*
 * Copyright 2004,2005 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.user.core.ldap;

import org.wso2.carbon.user.core.config.RealmConfiguration;
import org.wso2.carbon.user.core.claim.ClaimManager;
import org.wso2.carbon.user.core.profile.ProfileConfigurationManager;
import org.wso2.carbon.user.core.UserRealm;
import org.wso2.carbon.user.core.UserStoreException;
import org.wso2.carbon.user.core.tenant.Tenant;
import org.wso2.carbon.CarbonConstants;
import org.apache.axiom.om.util.Base64;

import javax.naming.directory.*;
import javax.naming.NamingException;
import javax.naming.NamingEnumeration;
import java.util.Map;
import java.util.Iterator;
import java.util.HashMap;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

public class ApacheDSUserStoreManager extends LDAPUserStoreManager {

    public static final String PASSWORD_HASH_METHOD = "passwordHashMethod";
    public static final String PASSWORD_HASH_METHOD_SHA = "SHA";
    public static final String PASSWORD_HASH_METHOD_MD5 = "MD5";

    public ApacheDSUserStoreManager(RealmConfiguration realmConfig,
                                    Map<String, Object> properties,
                                    ClaimManager claimManager,
                                    ProfileConfigurationManager profileManager,
                                    UserRealm realm,
                                    Integer tenantId) throws UserStoreException {
        super(realmConfig, properties, claimManager, profileManager, realm, tenantId);
    }

    public ApacheDSUserStoreManager(RealmConfiguration realmConfig,
                                    ClaimManager claimManager,
                                    ProfileConfigurationManager profileManager,
                                    String domainName) throws UserStoreException {
        super(realmConfig, claimManager, profileManager);
    }

    public void addUser(String userName, Object credential, String[] roleList,
                        Map<String, String> claims, String profileName) throws UserStoreException {
        this.addUser(userName, credential, roleList, claims, profileName, false);
    }

    public void addUser(String userName, Object credential, String[] roleList,
                        Map<String, String> claims, String profileName, boolean requirePasswordChange) throws UserStoreException {

        if (!checkUserNamePasswordValid(userName, credential)) {
            throw new UserStoreException("User name and password does not match");
        }

        DirContext dirContext = this.connectionSource.getContext();
        String searchBase = realmConfig.getUserStoreProperty(LDAPConstants.USER_SEARCH_BASE);
        try {

            dirContext = (DirContext) dirContext.lookup(searchBase);
            //above this the searchbase or base directoy for this storage. we assume users are stored under
            //ou=users folder here
            dirContext = (DirContext) dirContext.lookup("ou=users");

            BasicAttributes basicAttributes = new BasicAttributes(true);

            // set the objectClass type for schema
            BasicAttribute objectClass = new BasicAttribute("objectClass");
            objectClass.add("inetOrgPerson");
            objectClass.add("organizationalPerson");
            objectClass.add("person");
            objectClass.add("top");
            basicAttributes.put(objectClass);

            BasicAttribute uid = new BasicAttribute("uid");
            uid.add(userName);
            basicAttributes.put(uid);

            BasicAttribute userPassword = new BasicAttribute("userPassword");
            userPassword.add(getPasswordToStore((String) credential, this.realmConfig.getUserStoreProperty(PASSWORD_HASH_METHOD)));
            basicAttributes.put(userPassword);

            for (String role : roleList) {
                this.hybridRoleManager.addHybridRole(role, new String[]{userName});
            }

            BasicAttribute claim;

            for (String key : claims.keySet()) {
                claim = new BasicAttribute(key);
                claim.add(claims.get(key));
                basicAttributes.put(claim);
            }

            BasicAttribute isChangePassword = new BasicAttribute("requirePasswordChange");
            isChangePassword.add(requirePasswordChange);

            // for the moment set required attributes cn, sn as user names
            BasicAttribute cn = new BasicAttribute("cn");
            cn.add(userName);
            basicAttributes.put(cn);

            BasicAttribute sn = new BasicAttribute("sn");
            sn.add(userName);
            basicAttributes.put(sn);

            dirContext.bind("uid=" + userName, null, basicAttributes);
        } catch (NamingException e) {
            throw new UserStoreException("Can not access the directory context or user already exists in the system", e);
        }
    }

    private String getPasswordToStore(String password, String passwordHashMethod) throws UserStoreException {
        String passwordToStore = password;
        if (passwordHashMethod != null) {
            try {
                MessageDigest messageDigest = MessageDigest.getInstance(passwordHashMethod);
                byte[] digestValue = messageDigest.digest(password.getBytes());
                passwordToStore = "{" + passwordHashMethod + "}" + Base64.encode(digestValue);
            } catch (NoSuchAlgorithmException e) {
                throw new UserStoreException("Invalid hashMethod", e);
            }
        }
        return passwordToStore;
    }

    public void deleteUser(String userName) throws UserStoreException {

        if (realmConfig.getAdminUserName().equals(userName)) {
            throw new UserStoreException("Cannot delete admin user");
        }

        if (CarbonConstants.REGISTRY_ANONNYMOUS_USERNAME.equals(userName)) {
            throw new UserStoreException("Cannot delete anonymous user");
        }

        this.hybridRoleManager.deleteUser(userName);

        DirContext dirContext = this.connectionSource.getContext();

        //first search the existing user entry.
        String searchBase = realmConfig.getUserStoreProperty(LDAPConstants.USER_SEARCH_BASE);
        String searchFilter = realmConfig.getUserStoreProperty(LDAPConstants.USER_FILTER);
        searchFilter = searchFilter.replace("?", userName);

        SearchControls searchControls = new SearchControls();
        searchControls.setSearchScope(SearchControls.SUBTREE_SCOPE);

        try {
            NamingEnumeration<SearchResult> namingEnumeration = dirContext.search(searchBase, searchFilter, searchControls);
            // here we assume only one user
            // TODO: what to do if there are more than one user
            while (namingEnumeration.hasMore()) {
                SearchResult searchResult = namingEnumeration.next();
                String dnName = searchResult.getName();
                dirContext = (DirContext) dirContext.lookup(searchBase);
                dirContext.destroySubcontext(dnName);
            }

        } catch (NamingException e) {
            throw new UserStoreException("Can not access the directory serviece", e);
        }
    }

    public void updateCredential(String userName, Object newCredential, Object oldCredential)
            throws UserStoreException {
        DirContext dirContext = this.connectionSource.getContext();

        //first search the existing user entry.
        String searchBase = realmConfig.getUserStoreProperty(LDAPConstants.USER_SEARCH_BASE);
        String searchFilter = realmConfig.getUserStoreProperty(LDAPConstants.USER_FILTER);
        searchFilter = searchFilter.replace("?", userName);

        SearchControls searchControls = new SearchControls();
        searchControls.setSearchScope(SearchControls.SUBTREE_SCOPE);
        searchControls.setReturningAttributes(new String[]{"userPassword"});

        try {
            NamingEnumeration<SearchResult> namingEnumeration = dirContext.search(searchBase, searchFilter, searchControls);
            // here we assume only one user
            // TODO: what to do if there are more than one user
            while (namingEnumeration.hasMore()) {
                SearchResult searchResult = namingEnumeration.next();
                Attributes attributes = searchResult.getAttributes();
                Attribute userPassword = attributes.get("userPassword");

                String passwordHashMethod = null;
                // when admin changes other user passwords he do not have to provide
                // the old password.
                if (oldCredential != null) {
                    // here it is only possible to have one password, if there are more
                    // every one should match with the given old password
                    NamingEnumeration passwords = userPassword.getAll();
                    if (passwords.hasMore()) {
                        byte[] byteArray = (byte[]) passwords.next();
                        String password = new String(byteArray);

                        if (password.startsWith("{")) {
                            passwordHashMethod = password.substring(password.indexOf("{") + 1, password.indexOf("}"));
                        }

                        if (!password.equals(getPasswordToStore((String) oldCredential, passwordHashMethod))) {
                            throw new UserStoreException("Old password does not match");
                        }
                    }
                }


                String dnName = searchResult.getName();
                dirContext = (DirContext) dirContext.lookup(searchBase);

                Attribute passwordAttribute = new BasicAttribute("userPassword");
                passwordAttribute.add(getPasswordToStore((String) newCredential, passwordHashMethod));
                BasicAttributes basicAttributes = new BasicAttributes(true);
                basicAttributes.put(passwordAttribute);
                dirContext.modifyAttributes(dnName, DirContext.REPLACE_ATTRIBUTE, basicAttributes);
            }

            if (userName.equals(this.adminUserName)) {
                this.connectionSource.updateCredential((String) newCredential);
            }

        } catch (NamingException e) {
            throw new UserStoreException("Can not access the directory serviece", e);
        }
    }

    public Map<String, String> getProperties(Tenant tenant) throws UserStoreException {
        Map<String, String> existingProperties = this.realmConfig.getUserStoreProperties();
        String tenantSufix = getTenantSuffix(tenant.getDomain());
        String propertyName = null;
        Map<String, String> newProperties = new HashMap<String,String>();
        for (Iterator<String> iter = existingProperties.keySet().iterator(); iter.hasNext();) {
            propertyName = iter.next();
            if (propertyName.equals(LDAPConstants.USER_SEARCH_BASE)){
                newProperties.put(propertyName, tenantSufix);
            } else {
                newProperties.put(propertyName, existingProperties.get(propertyName));
            }
        }
        return newProperties;
    }

    private String getTenantSuffix(String domain) {
        // here we use a simple algorithum by splitting the domain with .
        String[] domainParts = domain.split("\\.");
        StringBuffer suffixName = new StringBuffer();
        for (String domainPart : domainParts) {
            suffixName.append(",dc=").append(domainPart);
        }
        return suffixName.toString().replaceFirst(",", "");
    }
}
