/*
 *  Copyright (c) 2005-2009, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 *  WSO2 Inc. licenses this file to you under the Apache License,
 *  Version 2.0 (the "License"); you may not use this file except
 *  in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *
 */
package org.wso2.carbon.user.core.listener;

import org.wso2.carbon.user.core.UserStoreException;
import org.wso2.carbon.user.core.AuthorizationManager;

public interface AuthorizationManagerListener {

    public static final int REGISTRY_AUTH_ROLE_LISTENER = 20;
    /**
     * execution id of the multi-tenancy restriction handler
     */
    public static final int MULTITENANCY_USER_RESTRICTION_HANDLER = 10;
    /**
     * execution id of the permission authorization listener
     */
    public static final int PERMISSION_AUTHORIZATION_LISTENER = 5;

    /**
     * Get the execution order identifier for this listener.
     * @return The execution order identifier integer value.
     */
    public int getExecutionOrderId();

    /**
     * Checks for user authorization.
     *
     * @param userName
     *            The user name
     * @param resourceId
     *            Resource identification string
     * @param action
     *            The action granted to the user
     * @param authorizationManager
     *            The underlying AuthorizationManager
     * @return Whether execution of this method of the underlying
     *         UserStoreManager must happen.
     * @throws UserStoreException
     *            Thrown by the underlying UserStoreManager
     */
    public boolean isUserAuthorized(String userName, String resourceId, String action,
                              AuthorizationManager authorizationManager) throws UserStoreException;

    /**
     * Checks for role authorization.
     *
     * @param roleName
     *            The role name
     * @param resourceId
     *            Resource identification string
     * @param action
     *            The action granted to the user
     * @param authorizationManager
     *            The underlying AuthorizationManager
     * @return Whether execution of this method of the underlying
     *         UserStoreManager must happen.
     * @throws UserStoreException
     *            Thrown by the underlying UserStoreManager
     */
    public boolean isRoleAuthorized(String roleName, String resourceId, String action,
                              AuthorizationManager authorizationManager) throws UserStoreException;

    /**
     * Grants authorizations to a user to perform an action on a resource.
     *
     * @param userName
     *            The user name
     * @param resourceId
     *            Resource identification string
     * @param action
     *            The action granted to the user
     * @param authorizationManager
     *            The underlying AuthorizationManager
     * @return Whether execution of this method of the underlying
     *         UserStoreManager must happen.
     * @throws UserStoreException
     *            Thrown by the underlying UserStoreManager
     */
    public boolean authorizeUser(String userName, String resourceId, String action,
                              AuthorizationManager authorizationManager) throws UserStoreException;

    /**
     * Grants authorizes to a role to perform an action on a resource.
     *
     * @param roleName
     *            The role name
     * @param resourceId
     *            Resource identification string
     * @param action
     *            The action granted to the role
     * @param authorizationManager
     *            The underlying AuthorizationManager
     * @return Whether execution of this method of the underlying
     *         UserStoreManager must happen.
     * @throws UserStoreException
     *            Thrown by the underlying UserStoreManager
     */
    public boolean authorizeRole(String roleName, String resourceId, String action,
                              AuthorizationManager authorizationManager) throws UserStoreException;

    /**
     * Deny authorizations to a user to perform an action on a resource.
     *
     * @param userName
     *            The user name
     * @param resourceId
     *            Resource identification string
     * @param action
     *            The action granted to the user
     * @param authorizationManager
     *            The underlying AuthorizationManager
     * @return Whether execution of this method of the underlying
     *         UserStoreManager must happen.
     * @throws UserStoreException
     *            Thrown by the underlying UserStoreManager
     */
    public boolean denyUser(String userName, String resourceId, String action,
                         AuthorizationManager authorizationManager) throws UserStoreException;

    /**
     * Deny authorizations to a role to perform an action on a resource.
     *
     * @param roleName
     *            The role name
     * @param resourceId
     *            Resource identification string
     * @param action
     *            The action granted to the role
     * @param authorizationManager
     *            The underlying AuthorizationManager
     * @return Whether execution of this method of the underlying
     *         UserStoreManager must happen.
     * @throws UserStoreException
     *            Thrown by the underlying UserStoreManager
     */
    public boolean denyRole(String roleName, String resourceId, String action,
                         AuthorizationManager authorizationManager) throws UserStoreException;

    /**
     * Deletes an already granted authorization of a user.
     *
     * @param userName
     *            The user name
     * @param resourceId
     *            Resource identification string
     * @param action
     *            The action granted
     * @param authorizationManager
     *            The underlying AuthorizationManager
     * @return Whether execution of this method of the underlying
     *         UserStoreManager must happen.
     * @throws UserStoreException
     *            Thrown by the underlying UserStoreManager
     */
    public boolean clearUserAuthorization(String userName, String resourceId, String action,
                                       AuthorizationManager authorizationManager) throws UserStoreException;

    /**
     *
     * @param userName the user name
     * @param authorizationManager
     *            The underlying AuthorizationManager
     * @return Whether execution of this method of the underlying
     *         UserStoreManager must happen.
     * @throws UserStoreException
     *            Thrown by the underlying UserStoreManager
     */
    public boolean clearUserAuthorization(String userName,
                                       AuthorizationManager authorizationManager) throws UserStoreException;

    /**
     * Deletes an already granted authorization of a role.
     *
     * @param roleName
     *            The role name
     * @param resourceId
     *            Resource identification string
     * @param action
     *            The action granted
     * @param authorizationManager
     *            The underlying AuthorizationManager
     * @return Whether execution of this method of the underlying
     *         UserStoreManager must happen.
     * @throws UserStoreException
     *            Thrown by the underlying UserStoreManager
     */
    public boolean clearRoleAuthorization(String roleName, String resourceId, String action,
                                       AuthorizationManager authorizationManager) throws UserStoreException;

    /**
     * Deletes the role's right to perform the action on all resources.
     *
     * @param roleName
     *            The role name
     * @param action
     *            The action granted
     * @param authorizationManager
     *            The underlying AuthorizationManager
     * @return Whether execution of this method of the underlying
     *         UserStoreManager must happen.
     * @throws UserStoreException
     *            Thrown by the underlying UserStoreManager
     */
    public boolean clearRoleActionOnAllResources(String roleName, String action,
                                              AuthorizationManager authorizationManager)
            throws UserStoreException;

    /**
     * Used when deleting roles.
     *
     * @param roleName
     *            The role name
     * @param authorizationManager
     *            The underlying AuthorizationManager
     * @return Whether execution of this method of the underlying
     *         UserStoreManager must happen.
     * @throws UserStoreException
     *            Thrown by the underlying UserStoreManager
     */
    public boolean clearRoleAuthorization(String roleName,
                                       AuthorizationManager authorizationManager) throws UserStoreException;

    /**
     * Deletes all granted authorization on a resource.
     *
     * @param resourceId
     *            Resource identification string
     * @param authorizationManager
     *            The underlying AuthorizationManager
     * @return Whether execution of this method of the underlying
     *         UserStoreManager must happen.
     * @throws UserStoreException
     *            Thrown by the underlying UserStoreManager
     */
    public boolean clearResourceAuthorizations(String resourceId,
                                            AuthorizationManager authorizationManager)
            throws UserStoreException;

}
